<?php

/* Icinga Dependencies Web | (c) 2024 Icinga GmbH | AGPLv3 */

namespace Icinga\Module\Dependencies\Controllers;

use GuzzleHttp\Psr7\Utils;
use Icinga\Application\Icinga;
use Icinga\Application\Logger;
use Icinga\Module\Dependencies\Widget\Map\DependencyNodeLabel;
use Icinga\Module\Dependencies\Widget\MapSettings;
use Icinga\Module\Icingadb\Common\Auth;
use Icinga\Module\Icingadb\Common\Backend;
use Icinga\Module\Icingadb\Common\Database;
use Icinga\Module\Icingadb\Common\SearchControls;
use Icinga\Module\Icingadb\Model\DependencyEdge;
use Icinga\Module\Icingadb\Model\DependencyEdgeState;
use Icinga\Module\Icingadb\Model\DependencyNode;
use Icinga\Module\Icingadb\Redis\VolatileStateResults;
use Icinga\Module\Icingadb\Web\Control\SearchBar\ObjectSuggestions;
use Icinga\Module\Icingadb\Web\Controller;
use Icinga\Module\Icingadb\Widget\ItemList\ObjectList;
use Icinga\Util\Json;
use Icinga\Web\FileCache;
use Icinga\Web\Session;
use ipl\Html\Attributes;
use ipl\Html\Form;
use ipl\Html\Html;
use ipl\Html\HtmlElement;
use ipl\Html\HtmlString;
use ipl\Sql\Adapter\Mysql;
use ipl\Sql\Adapter\Pgsql;
use ipl\Sql\Expression;
use ipl\Stdlib\Filter;
use ipl\Stdlib\Str;
use ipl\Web\Common\CsrfCounterMeasure;
use ipl\Web\Filter\QueryString;
use ipl\Web\Widget\EmptyStateBar;
use ipl\Web\Widget\Icon;
use ipl\Web\Widget\ItemList;
use JShrink\Minifier;

class MapController extends Controller
{
    use Auth;
    use Database;
    use SearchControls;
    use CsrfCounterMeasure;

    public function indexAction(): void
    {
        $this->addTitleTab($this->translate('Dependency Map'));

        if (! Backend::supportsDependencies()) {
            $this->addContent(new EmptyStateBar(
                $this->translate('Upgrade Icinga DB to v1.4 or later to use this feature')
            ));

            return;
        }

        $this->setAutorefreshInterval(60);
        $compact = $this->view->compact;

        $nodeIds = DependencyNode::on($this->getDb())
            ->columns('id')
            ->disableDefaultSort();
        $nodes = DependencyNode::on($this->getDb())
            ->with([
                'host',
                'host.state',
                'host.state.last_comment',
                'service',
                'service.state',
                'service.state.last_comment',
                'service.host',
                'service.host.state',
                'redundancy_group',
                'redundancy_group.state'
            ])
            ->limit(50)
            ->setResultSetClass(VolatileStateResults::class);

        $limitControl = $this->createLimitControl();
        $paginationControl = $this->createPaginationControl($nodes);
        $sortControl = $this->createSortControl($nodes, [
            'name'                                  => $this->translate('Name'),
            'severity desc, last_state_change desc' => $this->translate('Severity'),
            'state'                                 => $this->translate('Current State'),
            'last_state_change desc'                => $this->translate('Last State Change')
        ]);
        $lodControl = $this->createLodControl();
        $groupControl = $this->createGroupControl();
        $viewModeSwitcher = $this->createViewModeSwitcher($paginationControl, $limitControl);
        $viewModeSwitcher->addWrapper(Html::tag('div', ['class' => ['icinga-module', 'module-icingadb']]));
        $searchBar = $this->createSearchBar($nodes, [
            $limitControl->getLimitParam(),
            $sortControl->getSortParam(),
            $viewModeSwitcher->getViewModeParam(),
            'groupBy'
        ]);

        if ($searchBar->hasBeenSent() && ! $searchBar->isValid()) {
            if ($searchBar->hasBeenSubmitted()) {
                $filter = QueryString::fromString((string) $this->params)->parse();
            } else {
                $this->addControl($searchBar);
                $this->sendMultipartUpdate();
                return;
            }
        } else {
            $filter = $searchBar->getFilter();
        }

        $searchResultContainer = Html::tag('div', [
            'id' => 'search-results',
            'class' => ['icinga-module', 'module-icingadb']
        ]);

        if ($filter instanceof Filter\Chain && $filter->isEmpty()) {
            $searchResults = [];
        } else {
            $this->filter($nodeIds, $filter);
            $this->filter($nodes, $filter);

            if (! $compact) {
                $this->addControl($paginationControl);
                $searchResultContainer->addHtml(
                    (new ObjectList($nodes))
                        ->addAttributes(['id' => 'map-search-results-list'])
                        ->setViewMode($viewModeSwitcher->getViewMode())
                        ->on(ItemList::ON_ITEM_ADD, function ($item, $node) {
                            $item->add(Html::tag('button', [
                                'type' => 'button',
                                'class' => 'control-button',
                                'data-node-id' => $this->serializeId($node->id),
                                'title' => $this->translate('Show on map')
                            ], new Icon('arrows-to-circle')));
                        })
                );
            }

            $ids = $this->getDb()->fetchCol($nodeIds->assembleSelect());
            $searchResults = array_map([$this, 'serializeId'], $ids);
        }

        $this->addControl($sortControl);
        $this->addControl($limitControl);
        $this->addControl($viewModeSwitcher);
        $this->addControl($groupControl);
        $this->addControl($lodControl);
        $this->addControl($searchBar);

        if (! $searchBar->hasBeenSubmitted() && $searchBar->hasBeenSent()) {
            $this->addPart(HtmlString::create(Json::encode($searchResults)), 'Behavior:DependenciesMap');
            if (empty($searchResults)) {
                $this->controls->remove($paginationControl);
                $this->addPart(HtmlString::create(''), $paginationControl->getAttribute('id')->getValue());
            }

            if ($searchResultContainer->isEmpty()) {
                // This is only added because multipart content must not be empty…
                $searchResultContainer->addHtml(Html::tag('div'));
            }

            $this->sendMultipartUpdate();
            $this->addPart($searchResultContainer);
        } else {
            $mapContainer = Html::tag('div', [
                'id' => 'map-container',
                'data-csrf-token' => $this->getCsrfToken(),
                'data-group-by' => $groupControl->getGroupBy(),
                'data-search-results' => Json::encode($searchResults)
            ], [
                Html::tag('div', ['id' => 'map-impact'], [
                    Html::tag('div', ['class' => 'progress'], [
                        Html::tag('div', ['class' => 'bar'])
                    ])
                ])
            ]);

            if (
                $this->getRequest()->getHeader('Cache-Control') === 'no-cache'
                || $this->getRequest()->getHeader('Pragma') === 'no-cache'
            ) {
                $mapContainer->setAttribute('data-invalidate', true);
            }

            if (empty($searchResults)) {
                $this->controls->remove($paginationControl);
                $this->controls->insertBefore(Html::tag('div', $paginationControl->getAttributes()), $sortControl);
            }

            $this->addContent(Html::tag('div', ['class' => 'map-wrapper'], [
                $searchResultContainer,
                $mapContainer
            ]));
        }
    }

    protected function createLodControl(): HtmlElement
    {
        return Html::tag('div', ['id' => 'lod-control', 'hidden' => true], [
            Html::tag('label', [
                Html::wantHtml('LOD'),
                Html::tag('input', [
                    'type' => 'range',
                    'value' => 1,
                    'min' => 1,
                    'list' => 'lod-ticks',
                ])
            ]),
            Html::tag('datalist', ['id' => 'lod-ticks'], [
                Html::tag('option', ['value' => 1]),
                Html::tag('option', ['value' => 2]),
                Html::tag('option', ['value' => 3]),
            ])
        ]);
    }

    protected function createGroupControl(): MapSettings
    {
        $groupBy = $this->params->shift('groupBy');

        $groupControl = new MapSettings();
        $groupControl->setGroupBy($groupBy);
        $groupControl->on(Form::ON_SENT, function (MapSettings $form) {
            if (! $form->hasBeenSubmitted()) {
                $this->addPart(...$form->getPartUpdates());
            }
        });
        $groupControl->on(Form::ON_SUCCESS, function (MapSettings $form) {
            $this->redirectNow(
                $this->getRequest()
                    ->getUrl()
                    ->onlyWith([])
                    ->setParams(array_filter($form->getValues()))
            );
        });
        $groupControl->handleRequest($this->getServerRequest());

        return $groupControl;
    }

    public function suggestGroupAction(): void
    {
        $request = $this->getServerRequest();

        // Lalala… I know this is ugly, but I didn't want to adjust the completion script to support this
        $body = json_decode($request->getBody()->read(8192), true);
        $body['term']['type'] = 'column';
        $request = $request->withBody(Utils::streamFor(json_encode($body)));

        // TODO: Add type prefixes otherwise facts cannot be distinguished, it also solves the issue that an
        //       invalid fact cannot be edited if the new label is equal to the previous fact
        $suggestions = new ObjectSuggestions();
        $suggestions->setModel(DependencyNode::class);
        $suggestions->forRequest($request);
        $this->getDocument()->add($suggestions);
    }

    public function completeAction(): void
    {
        $suggestions = new ObjectSuggestions();
        $suggestions->setModel(DependencyNode::class);
        $suggestions->forRequest($this->getServerRequest());
        $this->getDocument()->add($suggestions);
    }

    private function stopOutputBuffering(): void
    {
        do {
            ob_end_clean();
        } while (ob_get_level() > 0);
    }

    private function serializeId($bin): string
    {
        if (is_resource($bin)) {
            // pdo_pgsql returns a resource
            $bin = stream_get_contents($bin);
        }

        return base64_encode($bin);
    }

    private function deserializeId($serial): ?string
    {
        return base64_decode($serial) ?: null;
    }

    private function getCsrfToken(?string $seed = null): string
    {
        $hashAlgo = in_array('sha3-256', hash_algos(), true) ? 'sha3-256' : 'sha256';
        $seed = $seed ?? random_bytes(16);

        return base64_encode($seed) . '|' . hash($hashAlgo, Session::getSession()->getId() . $seed);
    }

    private function validateCsrfToken(string $token): bool
    {
        return $this->createCsrfCounterMeasure(Session::getSession()->getId())->setValue($token)->isValid();
    }

    public function workerAction(): void
    {
        $jsDir = $this->Module()->getJsDir();
        $script = file_get_contents($jsDir . '/map/Worker.js');

        $this->getResponse()
            ->setHeader('Content-Type', 'application/javascript')
            ->setHeader('Cache-Control', 'no-store')
            ->sendResponse();

        $this->stopOutputBuffering();

        echo $script;

        exit;
    }

    public function workerDependenciesAction(): void
    {
        $jsDir = $this->Module()->getJsDir();
        $paths = [
            Icinga::app()
                ->getLibraries()
                ->get('icinga-php-library')
                ->getJsAssetPath() . '/iterator.js',
            "$jsDir/map/Utils.js",
            "$jsDir/map/GroupForce.js"
        ];
        if ($this->params->has('d3')) {
            array_unshift(
                $paths,
                Icinga::app()
                    ->getLibraries()
                    ->get('icinga-php-thirdparty')
                    ->getJsAssetPath() . '/mbostock/d3.min.js'
            );
        }

        $request = $this->getServerRequest();
        $response = $this->getResponse();
        $response->setHeader('Cache-Control', 'public,no-cache,must-revalidate');

        $noCache = in_array('no-cache', array_merge(
            $request->getHeader('Cache-Control'),
            $request->getHeader('Pragma')
        ), true);
        if (! $noCache && FileCache::etagMatchesFiles($paths)) {
            $response->setHeader('HTTP/1.1 304 Not Modified')
                ->sendResponse();

            exit;
        }

        $etag = FileCache::etagForFiles($paths);
        $response->setHeader('ETag', $etag);
        $response->setHeader('Content-Type', 'application/javascript');

        $cacheFile = 'worker-dependencies-' . $etag . '.js';
        $cache = FileCache::instance('dependencies');
        if ($noCache || ! $cache->has($cacheFile)) {
            $script = '';
            foreach ($paths as $scriptPath) {
                $script .= file_get_contents($scriptPath) . "\n\n\n";
            }

            $cache->store($cacheFile, Minifier::minify($script, ['flaggedComments' => false]));
        }

        $response->sendResponse();
        $this->stopOutputBuffering();
        $cache->send($cacheFile);

        exit;
    }

    public function structureAction(): void
    {
        $this->_helper->layout()->disableLayout();

        if ($this->getRequest()->isApiRequest()) {
            $this->httpBadRequest('Cannot respond with JSON');
        }

        $groupBy = $this->params->get('groupBy');
        $invalidate = (bool) $this->params->get('invalidate');
        $cacheFile = $this->cacheFile($groupBy);
        $cache = FileCache::instance('dependencies');
        if (! $invalidate && $cache->has($cacheFile)) {
            // This is a bit of a hack, but I don't want to deal with the cache's internals
            $filePath = sys_get_temp_dir() . '/FileCache_icingaweb/modules/dependencies/' . $cacheFile;

            $this->getResponse()
                ->setHeader('Content-Type', 'application/json')
                ->setHeader('Content-Length', filesize($filePath))
                ->setHeader('Cache-Control', 'no-store')
                ->sendResponse();

            $this->stopOutputBuffering();

            readfile($filePath);

            exit;
        }

        $groupBy = Str::trimSplit($groupBy);

        $nodes = DependencyNode::on($this->getDb())
            ->disableDefaultSort() // I want to put as less strain on the database as possible
            ->columns('id');

        $groupConcat = 'GROUP_CONCAT(%s SEPARATOR "|")';
        if ($this->getDb()->getAdapter() instanceof Pgsql) {
            $groupConcat = "ARRAY_TO_STRING(ARRAY_AGG(%s), '|')";
        }

        //$hostVarColumns = [];
        $serviceIdUsed = false;
        foreach ($groupBy as $i => $column) {
            $groupColumn = 'redundancy_group.from.to.' . $column;
            if (substr($column, 4, 6) === '.vars.' || substr($column, 7, 6) === '.vars.') {
                /*if (substr($column, 0, 5) === 'host.') {
                    $hostVarColumns["g_$i"] = null;
                    $hostVarColumns["rg_$i"] = null;
                }*/
            } elseif (substr($column, 0, 5) === 'host.' || substr($column, 0, 10) === 'hostgroup.') {
                $column = new Expression(
                    "CASE WHEN %s IS NULL THEN $groupConcat ELSE NULL END",
                    ['service_id', $column]
                );
                $groupColumn = new Expression(
                    "CASE WHEN %s IS NULL THEN $groupConcat ELSE NULL END",
                    ['redundancy_group.from.to.service_id', $groupColumn]
                );
                $serviceIdUsed = true;
            } else {
                $column = new Expression($groupConcat, [$column]);
                $groupColumn = new Expression($groupConcat, [$groupColumn]);
            }

            $nodes->withColumns([
                "g_$i" => $column,
                "rg_$i" => $groupColumn
            ]);
        }

        $this->applyRestrictions($nodes);

        $nodeSelect = $nodes->assembleSelect();

        if ($groupBy) {
            $nodeSelect->groupBy('dependency_node.id');
            if ($serviceIdUsed) {
                $resolver = $nodes->getResolver();
                $nodeSelect->groupBy([
                    $resolver->qualifyColumn('service_id', $resolver->getAlias($nodes->getModel())),
                    $resolver->qualifyColumn('service_id', $resolver->getAlias(
                        $resolver->resolveRelation('dependency_node.redundancy_group.from.to')->getTarget()
                    ))
                ]);
            }

            // TODO: Decide if we want to strictly differentiate between host and service vars
            /*foreach (array_intersect_key($nodes->getColumns(), $hostVarColumns) as $name => $expression) {
                $nodes->columns([$name => new Expression(
                    str_replace(
                        '= dependency_node.id',
                        '= dependency_node.id AND dependency_node.service_id IS NULL',
                        $expression->getStatement()
                    ),
                    $expression->getColumns(),
                    ...$expression->getValues()
                )]);
            }*/

            // TODO: Should be the orm's job, but at the moment relation join types are static
            $join = $nodeSelect->getJoin();
            $nodeSelect->resetJoin();
            foreach ($join ?? [] as $j) {
                $nodeSelect->joinLeft($j[1], $j[2]);
            }
        }

        $edges = DependencyEdge::on($this->getDb())
            ->disableDefaultSort()
            ->columns([
                'dependency_edge_state_id',
                'from_node_id',
                'to_node_id'
            ])->assembleSelect();

        $this->getResponse()
            ->setHeader('Content-Type', 'text/plain')
            ->setHeader('Cache-Control', 'no-store')
            ->sendResponse();

        try {
            $this->stopOutputBuffering();

            $batchProcessing = true;
            if ($this->getDb()->getAdapter() instanceof Mysql) {
                $this->getDb()->setAttribute(\PDO::MYSQL_ATTR_USE_BUFFERED_QUERY, false);
                $batchProcessing = false;
            } else {
                $nodeSelect->orderBy('id')->limit(1000);
            }

            $offset = 0;

            do {
                $nodeSelect->offset($offset);
                $i = 0;
                $identifiedGroups = [];
                foreach ($this->getDb()->select($nodeSelect) as $i => $row) {
                    $groups = [];
                    for ($j = 0; $j < count($groupBy); $j++) {
                        $groupByValue = $row->{"g_$j"} ?? $row->{"rg_$j"};
                        if ($groupByValue) {
                            $groups[] = $groupByValue;
                        }
                    }

                    $group = implode('|', array_unique($groups));
                    $identifiedGroups[$group] = true;

                    $data = [1, $this->serializeId($row->id)];
                    if ($group) {
                        $data[] = $group;
                    }

                    echo Json::sanitize($data), "\n";
                }

                // TODO: I'm not entirely sure whether this is the right threshold, let's wait for user feedback
                if ($groupBy && count($identifiedGroups) / $i > 0.25) {
                    echo Json::sanitize([
                        3,
                        $this->translate('Given group facts are not unique enough. Please choose different facts.'),
                        (new EmptyStateBar(
                            $this->translate('Given group facts are not unique enough. Please choose different facts.')
                        ))->render()
                    ]), "\n";

                    exit;
                } elseif ($i === 0 && $offset === 0) {
                    echo Json::sanitize([
                        3,
                        $this->translate('No dependency nodes found'),
                        (new EmptyStateBar($this->translate('No dependency nodes found')))->render()
                    ]), "\n";

                    exit;
                }

                $offset += 1000;
            } while ($i === 999);

            if ($batchProcessing) {
                $edges->orderBy('id')->limit(1000);
            }

            $offset = 0;

            do {
                $edges->offset($offset);
                $i = 0;
                foreach ($this->getDb()->select($edges) as $i => $row) {
                    $data = [
                        2,
                        $offset++,
                        $this->serializeId($row->dependency_edge_state_id),
                        $this->serializeId($row->from_node_id),
                        $this->serializeId($row->to_node_id)
                    ];

                    echo Json::sanitize($data), "\n";
                }
            } while ($i === 999);
        } catch (\Throwable $e) {
            Logger::error('Fatal error while fetching dependency map structure: %s', $e->getMessage());
            Logger::debug($e->getTraceAsString());

            echo Json::sanitize([
                3,
                $this->translate('Fatal error while fetching dependency map structure'),
                (new EmptyStateBar($this->translate(
                    'Fatal error while fetching dependency map structure. Check log for further details'
                )))->render()
            ]), "\n";
        }

        exit;
    }

    private function cacheFile($groupBy)
    {
        $roles = [];
        foreach ($this->Auth()->getUser()->getRoles() as $role) {
            $roles[] = $role->getName();
        }

        return 'map-structure-' . hash('sha256', $groupBy . implode('', $roles)) . '.json';
    }

    public function cacheAction(): void
    {
        $this->assertHttpMethod('POST');
        $groupBy = $this->params->shift('groupBy');
        $csrfToken = $this->params->shiftRequired('csrfToken');
        if (! $this->validateCsrfToken($csrfToken)) {
            $this->httpBadRequest('Invalid CSRF token');
        }

        $data = $this->getServerRequest()->getBody()->getContents();

        FileCache::instance('dependencies')
            ->store($this->cacheFile($groupBy), $data);

        exit;
    }

    public function statesAction(): void
    {
        $this->_helper->layout()->disableLayout();

        if ($this->getRequest()->isApiRequest()) {
            $this->httpBadRequest('Cannot respond with JSON');
        }


        $nodes = DependencyNode::on($this->getDb())
            ->disableDefaultSort()
            ->columns([
                'id',
                'type' => new Expression(
                    "CASE
                        WHEN %s IS NOT NULL THEN 'service'
                        WHEN %s IS NOT NULL THEN 'host'
                        ELSE 'redundancy_group'
                    END",
                    [
                        'service_id',
                        'host_id'
                    ]
                ),
                'state' => new Expression(
                    "CASE
                        WHEN %s IS NOT NULL THEN %s
                        WHEN %s IS NOT NULL THEN %s
                        ELSE CASE WHEN %s = 'y' THEN 1 ELSE 0 END
                    END",
                    [
                        'service_id',
                        'service.state.soft_state',
                        'host_id',
                        'host.state.soft_state',
                        'redundancy_group.state.failed'
                    ]
                )
            ]);

        $this->applyRestrictions($nodes);

        $nodes = $nodes->assembleSelect();

        $edges = DependencyEdgeState::on($this->getDb())
            ->disableDefaultSort()
            ->columns([
                'id',
                'failed' => new Expression("CASE WHEN %s = 'y' THEN 1 ELSE 0 END", ['failed']),
                'total_children' => new Expression("COUNT(%s)", ['edge.from_node_id']),
            ]);

        $edges = $edges->assembleSelect()->groupBy(['dependency_edge_state.id', 'dependency_edge_state.failed']);

        $this->getResponse()
            ->setHeader('Content-Type', 'text/plain')
            ->setHeader('Cache-Control', 'no-store')
            ->sendResponse();

        try {
            $this->stopOutputBuffering();

            if ($this->getDb()->getAdapter() instanceof Mysql) {
                $this->getDb()->setAttribute(\PDO::MYSQL_ATTR_USE_BUFFERED_QUERY, false);
            } else {
                $nodes->orderBy('id')->limit(1000);
                $edges->orderBy('id')->limit(1000);
            }

            $offset = 0;

            do {
                $nodes->offset($offset);
                $i = 0;
                foreach ($this->getDb()->select($nodes) as $i => $row) {
                    $serial = $this->serializeId($row->id);
                    if ($serial === null) {
                        continue;
                    }

                    echo Json::sanitize([1, $serial, $row->type, $row->state]), "\n";
                }

                $offset += 1000;
            } while ($i === 999);

            $offset = 0;

            do {
                $edges->offset($offset);
                $i = 0;
                foreach ($this->getDb()->select($edges) as $i => $row) {
                    $serial = $this->serializeId($row->id);
                    if ($serial === null) {
                        continue;
                    }

                    echo Json::sanitize([2, $serial, $row->failed, $row->total_children]), "\n";
                }

                $offset += 1000;
            } while ($i === 999);
        } catch (\Throwable $e) {
            Logger::error('Fatal error while fetching dependency map states: %s', $e->getMessage());
            Logger::debug($e->getTraceAsString());

            echo Json::sanitize([
                3,
                $this->translate('Fatal error while fetching dependency map states'),
                (new EmptyStateBar($this->translate(
                    'Fatal error while fetching dependency map states. Check log for further details'
                )))->render()
            ]), "\n";
        }

        exit;
    }

    public function nodeLabelAction(): void
    {
        $this->_helper->layout()->disableLayout();

        $realId = $this->deserializeId($this->params->getRequired('id'));
        if ($realId === null) {
            $this->httpNotFound('Unknown Node');
        }

        $nodeQuery = DependencyNode::on($this->getDb())
            ->with([
                'host',
                'host.state',
                'host.icon_image',
                'service',
                'service.state',
                'service.icon_image',
                'service.host',
                'service.host.state',
                'redundancy_group',
                'redundancy_group.state'
            ])
            ->setResultSetClass(VolatileStateResults::class)
            ->filter(Filter::equal('id', $realId));
        $this->applyRestrictions($nodeQuery);

        /** @var DependencyNode $node */
        $node = $nodeQuery->first();
        if ($node === null) {
            $this->httpNotFound('Unknown Node');
        }

        $this->getResponse()
            ->setHeader('Content-Type', 'text/html')
            ->setHeader('Cache-Control', 'no-store')
            ->sendResponse();

        $this->stopOutputBuffering();

        echo (new HtmlElement(
            'div',
            Attributes::create(['class' => ['icinga-module', 'module-icingadb']]),
            new DependencyNodeLabel($node)
        ))->render();

        exit;
    }
}
