<?php

/* Icinga Dependencies Web | (c) 2025 Icinga GmbH | AGPLv3 */

namespace Icinga\Module\Dependencies\Common;

use Icinga\Module\Dependencies\Model\UnreachableParentWithPath;
use Icinga\Module\Icingadb\Common\Auth;
use Icinga\Module\Icingadb\Common\Backend;
use Icinga\Module\Icingadb\Model\Host;
use Icinga\Module\Icingadb\Model\RedundancyGroup;
use Icinga\Module\Icingadb\Model\Service;
use ipl\Orm\Model;
use ipl\Orm\Query;
use ipl\Sql\Select;

trait RootProblemPathQueryUtils
{
    use Auth;

    /**
     * Create the query for root problems with path
     *
     * @param Host|Service|RedundancyGroup $object
     *
     * @return Query
     */
    private function createQuery(Model $object): Query
    {
        // The UnreachableParentWithPath fetches all nodes (root problems and paths with ids to create path chain).
        // Applying restrictions can lead to missing chain nodes, because non-restricted nodes are filtered out, which
        // in turn leads to a broken chain. That's why the restrictions must be applied to Host and Service only, so we
        // still can fetch all related ids to create the chain.
        // Don't add host/service relations to with(). Instead create subqueries, apply restrictions and join
        // them to this query.
        $rootProblems = UnreachableParentWithPath::on(Backend::getDb(), $object)
            ->with(['redundancy_group', 'redundancy_group.state']);

        $hostsQuery = Host::on(Backend::getDb())
            ->with(['state', 'icon_image', 'state.last_comment']);
        $this->applyRestrictions($hostsQuery);

        $servicesQuery = Service::on(Backend::getDb())
            ->with(['state', 'icon_image', 'state.last_comment', 'host', 'host.state']);
        $this->applyRestrictions($servicesQuery);

        $prefix = $rootProblems->getModel()->getTableName() . '_';
        $rootProblems->getSelectBase()
            ->columns([$prefix . 'host.*', $prefix . 'service.*']) // add all subquery columns
            ->joinLeft(
                [$prefix . 'host' => $this->addColumnAliases($hostsQuery->assembleSelect(), $prefix)],
                $this->createJoinCondition($rootProblems, $hostsQuery)
            )
            ->joinLeft(
                [$prefix . 'service' => $this->addColumnAliases($servicesQuery->assembleSelect(), $prefix)],
                $this->createJoinCondition($rootProblems, $servicesQuery)
            )->orderBy([
                $prefix . 'host_state_severity', //column aliases
                $prefix . 'host_state_last_state_change',
                $prefix . 'service_state_severity',
                $prefix . 'service_state_last_state_change',
                $prefix . 'redundancy_group_state_failed',
                $prefix . 'redundancy_group_state_last_state_change'
            ], SORT_DESC);


        $rootProblems->on($rootProblems::ON_SELECT_ASSEMBLED, function (Select $select) use ($rootProblems) {
            // Select is assembled, the with() will not join relations anymore, but the hydrator can access them.
            // Add all host/service relations to avoid that the hydrator join tables for already fetched columns.
            // E.g. host.checkcommand_name column is fetched by the subquery with alias and added to the root query,
            // no need to join it as host.checkcommand.name
            $rootProblems->with([
                'host',
                'host.state',
                'host.icon_image',
                'host.state.last_comment',
                'service',
                'service.state',
                'service.icon_image',
                'service.state.last_comment',
                'service.host',
                'service.host.state'
            ]);
        });

        return $rootProblems;
    }

    /**
     * Create the `candidate key` (root query) and the `target key` (subquery) using given queries
     *
     * @param Query $rootQuery The root query to create candidate key
     * @param Query $subQuery The subquery to create target key
     *
     * @return string candidate key = target key
     */
    private function createJoinCondition(Query $rootQuery, Query $subQuery): string
    {
        $rootQueryResolver = $rootQuery->getResolver();
        $subQueryResolver = $subQuery->getResolver();

        $subQueryResolver->setAliasPrefix($rootQuery->getModel()->getTableName() . '_');

        $candidateKey = $rootQueryResolver->qualifyColumn('id', $rootQueryResolver->getAlias($rootQuery->getModel()));
        $targetKey = $subQueryResolver->qualifyColumnAlias('id', $subQueryResolver->getAlias($subQuery->getModel()));

        return $candidateKey . ' = ' . $targetKey;
    }

    /**
     * Add column aliases with given prefix to the subquery
     *
     * The subquery columns are unknown to the ORM so the ORM will create joins for these.
     * To prevent this behavior, the column aliases are applied in the same way as the ORM would have done.
     *
     * @param Select $select The select to prefix the columns
     * @param string $prefix The prefix to use
     *
     * @return Select
     */
    private function addColumnAliases(Select $select, string $prefix): Select
    {
        $columns = [];
        foreach ($select->getColumns() as $alias => $column) {
            if (is_int($alias)) {
                $columns[$prefix . str_replace('.', '_', $column)] = $column;
            } else {
                $columns[$prefix . $alias] = $column;
            }
        }

        return $select->resetColumns()->columns($columns);
    }
}
