<?php

/* Icinga Dependencies Web | (c) 2024 Icinga GmbH | AGPLv3 */

namespace Icinga\Module\Dependencies\ProvidedHook\Icingadb;

use Icinga\Module\Dependencies\Widget\Detail\ParentsAndChildren;
use Icinga\Module\Icingadb\Common\Auth;
use Icinga\Module\Icingadb\Common\Backend;
use Icinga\Module\Icingadb\Common\Database;
use Icinga\Module\Icingadb\Hook\HostDetailExtensionHook;
use Icinga\Module\Icingadb\Model\DependencyEdge;
use Icinga\Module\Icingadb\Model\DependencyNode;
use Icinga\Module\Icingadb\Model\Host;
use Icinga\Module\Icingadb\Redis\VolatileStateResults;
use ipl\Html\Attributes;
use ipl\Html\HtmlDocument;
use ipl\Html\HtmlElement;
use ipl\Html\Text;
use ipl\Html\ValidHtml;
use ipl\I18n\Translation;
use ipl\Orm\Query;
use ipl\Sql\Expression;
use ipl\Sql\Filter\Exists;
use ipl\Stdlib\Filter;
use ipl\Web\Url;
use ipl\Web\Widget\Icon;
use ipl\Web\Widget\Link;

class HostParentsAndChildrenExtension extends HostDetailExtensionHook
{
    use Auth;
    use Database;
    use Translation;

    protected function init(): void
    {
        $this->setLocation(510);
    }

    public function getHtmlForObject(Host $host): ValidHtml
    {
        if (! Backend::supportsDependencies()) {
            return new HtmlDocument();
        }

        $relations = [
            'host',
            'host.icon_image',
            'host.state',
            'service',
            'service.icon_image',
            'service.state',
            'service.host',
            'service.host.state',
            'redundancy_group',
            'redundancy_group.state'
        ];

        $sortOrder = [
            'host.state.severity',
            'host.state.last_state_change',
            'service.state.severity',
            'service.state.last_state_change',
            'redundancy_group.state.failed',
            'redundancy_group.state.last_state_change'
        ];

        $parentsQuery = DependencyNode::on($this->getDb())
            ->with($relations)
            ->orderBy($sortOrder, SORT_DESC)
            ->setResultSetClass(VolatileStateResults::class)
            ->limit(3);

        $this->filter($parentsQuery, $host->id, true);

        $this->applyRestrictions($parentsQuery);

        $childrenQuery = DependencyNode::on($this->getDb())
            ->with($relations)
            ->orderBy($sortOrder, SORT_DESC)
            ->setResultSetClass(VolatileStateResults::class)
            ->limit(3);

        $this->filter($childrenQuery, $host->id);

        $this->applyRestrictions($childrenQuery);

        $parents = $parentsQuery->execute();
        $children = $childrenQuery->execute();

        if (! $parents->hasResult() && ! $children->hasResult()) {
            return new HtmlDocument();
        }

        return (new HtmlDocument())
            ->addHtml(
                new HtmlElement(
                    'div',
                    Attributes::create(['class' => 'parents-and-children-header']),
                    new HtmlElement('h2', null, Text::create($this->translate('Dependencies'))),
                    (new Link(
                        [
                            new Icon('map-location-dot'),
                            new Text($this->translate('Show on map'))
                        ],
                        Url::fromPath('dependencies/map')->setFilter(Filter::all(
                            Filter::equal('host.name', $host->name),
                            Filter::unlike('service.name', '*')
                        )),
                        Attributes::create([
                            'title' => $this->translate('Show this host on the map'),
                            'class' => ['control-button', 'map-link']
                        ])
                    ))->setBaseTarget('_next')
                ),
                new ParentsAndChildren($host, $parents, $children)
            );
    }

    /**
     * Filter the query to only include (direct) parents or children of the given object.
     *
     * @todo This is a workaround, remove it once https://github.com/Icinga/ipl-orm/issues/76 is fixed
     *
     * @param Query $query
     * @param string $objectId
     * @param bool $fetchParents Fetch parents if true, children otherwise
     */
    private function filter(Query $query, string $objectId, bool $fetchParents = false): void
    {
        $filterTable = $fetchParents ? 'child' : 'parent';
        $utilizeType = $fetchParents ? 'parent' : 'child';

        $edge = DependencyEdge::on($this->getDb())
            ->utilize($utilizeType)
            ->columns([new Expression('1')])
            ->filter(Filter::equal("$filterTable.host.id", $objectId))
            ->filter(Filter::unlike("$filterTable.service.id", '*'));

        $edge->getFilter()->metaData()->set('forceOptimization', false);

        $resolver = $edge->getResolver();

        $edgeAlias = $resolver->getAlias(
            $resolver->resolveRelation($resolver->qualifyPath($utilizeType, $edge->getModel()->getTableName()))
                ->getTarget()
        );

        $query->filter(new Exists(
            $edge->assembleSelect()
                ->where(
                    "$edgeAlias.id = "
                    . $query->getResolver()->qualifyColumn('id', $query->getModel()->getTableName())
                )
        ));
    }
}
