<?php

/* Icinga Dependencies Web | (c) 2024 Icinga GmbH | AGPLv3 */

namespace Icinga\Module\Dependencies\ProvidedHook\Icingadb;

use Icinga\Module\Dependencies\Widget\Detail\ParentsAndChildren;
use Icinga\Module\Icingadb\Common\Auth;
use Icinga\Module\Icingadb\Common\Database;
use Icinga\Module\Icingadb\Hook\RedundancyGroupDetailExtensionHook;
use Icinga\Module\Icingadb\Model\DependencyNode;
use Icinga\Module\Icingadb\Model\RedundancyGroup;
use Icinga\Module\Icingadb\Redis\VolatileStateResults;
use ipl\Html\Attributes;
use ipl\Html\HtmlDocument;
use ipl\Html\HtmlElement;
use ipl\Html\Text;
use ipl\Html\ValidHtml;
use ipl\I18n\Translation;
use ipl\Stdlib\Filter;
use ipl\Web\Url;
use ipl\Web\Widget\Icon;
use ipl\Web\Widget\Link;

class RedundancyGroupParentsAndChildrenExtension extends RedundancyGroupDetailExtensionHook
{
    use Auth;
    use Database;
    use Translation;

    protected function init(): void
    {
        $this->setLocation(510);
    }

    public function getHtmlForObject(RedundancyGroup $redundancyGroup): ValidHtml
    {
        $relations = [
            'host',
            'host.icon_image',
            'host.state',
            'service',
            'service.icon_image',
            'service.state',
            'service.host',
            'service.host.state'
        ];

        $sortOrder = [
            'host.state.severity',
            'host.state.last_state_change',
            'service.state.severity',
            'service.state.last_state_change'
        ];

        $parentsQuery = DependencyNode::on($this->getDb())
            ->with($relations)
            ->filter(Filter::equal('child.redundancy_group.id', $redundancyGroup->id))
            ->orderBy($sortOrder, SORT_DESC)
            ->setResultSetClass(VolatileStateResults::class)
            ->limit(3);

        $this->applyRestrictions($parentsQuery);

        $childrenQuery = DependencyNode::on($this->getDb())
            ->with($relations)
            ->filter(Filter::equal('parent.redundancy_group.id', $redundancyGroup->id))
            ->orderBy($sortOrder, SORT_DESC)
            ->setResultSetClass(VolatileStateResults::class)
            ->limit(3);

        $this->applyRestrictions($childrenQuery);

        $parents = $parentsQuery->execute();
        $children = $childrenQuery->execute();

        if (! $parents->hasResult() && ! $children->hasResult()) {
            return new HtmlDocument();
        }

        return (new HtmlDocument())
            ->addHtml(
                new HtmlElement(
                    'div',
                    Attributes::create(['class' => 'parents-and-children-header']),
                    new HtmlElement('h2', null, Text::create($this->translate('Dependencies'))),
                    (new Link(
                        [
                            new Icon('map-location-dot'),
                            new Text($this->translate('Show on map'))
                        ],
                        Url::fromPath('dependencies/map')->setFilter(
                            Filter::equal('redundancy_group.id', bin2hex($redundancyGroup->id))
                        ),
                        Attributes::create([
                            'title' => $this->translate('Show this redundancy group on the map'),
                            'class' => ['control-button', 'map-link']
                        ])
                    ))->setBaseTarget('_next')
                ),
                new ParentsAndChildren($redundancyGroup, $parents, $children)
            );
    }
}
