<?php

/* Icinga Dependencies Web | (c) 2024 Icinga GmbH | AGPLv3 */

namespace Icinga\Module\Dependencies\View;

use Icinga\Module\Dependencies\Data\UnreachableParentResults;
use Icinga\Module\Icingadb\Common\Icons;
use Icinga\Module\Icingadb\Model\DependencyNode;
use Icinga\Module\Icingadb\Model\Host;
use Icinga\Module\Icingadb\Model\RedundancyGroup;
use Icinga\Module\Icingadb\Model\Service;
use ipl\Html\Attributes;
use ipl\Html\Html;
use ipl\Html\HtmlDocument;
use ipl\Html\HtmlElement;
use ipl\Html\Text;
use ipl\Orm\Model;
use ipl\Web\Widget\Icon;

/**
 * Renders the path to the root problem in the footer
 */
trait RootProblemPath
{
    /** @var UnreachableParentResults */
    protected $data;

    /** @var Host|Service|RedundancyGroup The current object */
    protected $currentObject;

    /** @var int The maximum path items to show */
    protected static $maxPathItems = 5;

    /**
     * Set the data
     *
     * @param UnreachableParentResults $data
     *
     * @return $this
     */
    public function setData(UnreachableParentResults $data): self
    {
        $this->data = $data;

        return $this;
    }

    /**
     * Set The current object
     *
     * @param Host|Service|RedundancyGroup $currentObject
     *
     * @return $this
     */
    public function setCurrentObject(Model $currentObject): self
    {
        $this->currentObject = $currentObject;

        return $this;
    }

    /**
     * Render the given object
     *
     * @param DependencyNode $object
     *
     * @return HtmlElement
     */
    private function renderItem(DependencyNode $object): HtmlElement
    {
        if ($object->service->id !== null) {
            $title = sprintf(
                $this->translate('%s on %s', '<service> on <host>'),
                $object->service->display_name,
                $object->service->host->display_name
            );

            $subject = Html::sprintf(
                $this->translate('%s on %s', '<service> on <host>'),
                new HtmlElement(
                    'span',
                    new Attributes(['class' => 'subject']),
                    new Text($object->service->display_name)
                ),
                new HtmlElement(
                    'span',
                    new Attributes(['class' => 'subject']),
                    new Text($object->service->host->display_name)
                )
            );
        } elseif ($object->host->id !== null || $object->redundancy_group->id !== null) {
            $title = $object->host->display_name ?? $object->redundancy_group->display_name;
            $subject = new HtmlElement('span', new Attributes(['class' => 'subject']), new Text($title));
        } else {
            $title = $this->translate('You are not authorized to view this object');
            $subject = new HtmlElement(
                'span',
                new Attributes(['class' => ['subject', 'unauthorized']]),
                new Text('***')
            );
        }

        return new HtmlElement(
            'div',
            new Attributes(['class' => ['item-badge'], 'title' => $title]),
            new Icon(Icons::UNREACHABLE),
            new HtmlElement('div', new Attributes(['class' => 'path']), $subject)
        );
    }

    public function assembleFooter($item, HtmlDocument $footer, string $layout): void
    {
        if (! isset($this->data, $this->currentObject)) {
            return;
        }

        $paths = $this->data->getPaths($item->id);

        while (! empty($paths)) {
            $pathItems = [];
            $popupItems = [];
            foreach (array_pop($paths) as $i => $pathItem) {
                $item = $this->renderItem($pathItem);
                if ($i < self::$maxPathItems) {
                    $pathItems[] = $item->addAttributes(['class' => 'path-item']);
                } else {
                    $popupItems[] = $item;
                }
            }

            if (empty($pathItems)) {
                return;
            }

            $allItems = $pathItems;

            if (! empty($popupItems)) {
                $allItems[] = new HtmlElement(
                    'div',
                    new Attributes(['class' => ['counter', 'item-badge']]),
                    new Text(sprintf('+ %d', count($popupItems))),
                    new HtmlElement('div', new Attributes(['class' => 'popup']), ...$popupItems)
                );
            }

            if ($this->currentObject instanceof Service) {
                $title = sprintf(
                    $this->translate('%s on %s', '<service> on <host>'),
                    $this->currentObject->display_name,
                    $this->currentObject->host->display_name
                );
            } else {
                $title = $this->currentObject->display_name;
            }

            $allItems[] = new HtmlElement(
                'div',
                new Attributes(['class' => ['item-badge', 'current-item'], 'title' => $title]),
                new Icon(Icons::UNREACHABLE)
            );

            $footer->addHtml(new HtmlElement('div', new Attributes(['class' => 'root-problem-path']), ...$allItems));
        }
    }
}
