<?php

/* Icinga Dependencies Web | (c) 2024 Icinga GmbH | AGPLv3 */

namespace Icinga\Module\Dependencies\Widget\Detail;

use Icinga\Module\Dependencies\Model\DependencyNodeSummary;
use Icinga\Module\Icingadb\Common\Auth;
use Icinga\Module\Icingadb\Common\Database;
use Icinga\Module\Icingadb\Common\Links;
use Icinga\Module\Icingadb\Model\Host;
use Icinga\Module\Icingadb\Model\RedundancyGroup;
use Icinga\Module\Icingadb\Model\Service;
use Icinga\Module\Icingadb\Widget\DependencyNodeStatistics;
use Icinga\Module\Icingadb\Widget\IconImage;
use ipl\Html\Attributes;
use ipl\Html\BaseHtmlElement;
use ipl\Html\Html;
use ipl\Html\HtmlDocument;
use ipl\Html\HtmlElement;
use ipl\Html\Text;
use ipl\Html\ValidHtml;
use ipl\Orm\Model;
use ipl\Orm\ResultSet;
use ipl\Stdlib\Filter;
use ipl\Web\Url;
use ipl\Web\Widget\Link;
use ipl\Web\Widget\StateBall;

class ParentsAndChildren extends BaseHtmlElement
{
    use Database;
    use Auth;

    protected $tag = 'div';

    protected $defaultAttributes = ['class' => 'parents-and-children'];

    /** @var Host|Service|RedundancyGroup The current object whose parents and children should be shown */
    protected $currentObject;

    /** @var ResultSet The result set for parents */
    protected $parents;

    /** @var ResultSet The result set for children */
    protected $children;

    /**
     * Create ParentsAndChildren nodes
     *
     * @param Host|Service|RedundancyGroup $currentObject The object whose parents and children should be shown
     * @param ResultSet $parents The result set for parents
     * @param ResultSet $children The result set for children
     */
    public function __construct(Model $currentObject, ResultSet $parents, ResultSet $children)
    {
        $this->currentObject = $currentObject;
        $this->parents = $parents;
        $this->children = $children;
    }

    protected function assemble(): void
    {
        $parentList = $this->createNodeList(true);
        $childList = $this->createNodeList();

        $this->addHtml(
            new HtmlElement(
                'div',
                Attributes::create(['class' => 'parents-label']),
                new HtmlElement('div', Attributes::create(['class' => 'label']), Text::create(t('Parents')))
            ),
            new HtmlElement(
                'div',
                Attributes::create(['class' => 'children-label']),
                new HtmlElement('div', Attributes::create(['class' => 'label']), Text::create(t('Children')))
            ),
            new HtmlElement(
                'div',
                Attributes::create(['class' => 'parents']),
                $parentList,
                $this->createSVGLines($parentList->count(), true)
            ),
            new HtmlElement(
                'div',
                Attributes::create(['class' => 'current']),
                new HtmlElement(
                    'span',
                    Attributes::create(['class' => 'self']),
                    $this->renderNode($this->currentObject)
                )
            ),
            new HtmlElement(
                'div',
                Attributes::create(['class' => 'children']),
                $this->createSVGLines($childList->count()),
                $childList
            )
        );
    }

    /**
     * Create the node list
     *
     * @param bool $forParents Whether the list is for parents or children
     *
     * @return HtmlElement
     */
    private function createNodeList(bool $forParents = false): HtmlElement
    {
        $list = new HtmlElement('ul', Attributes::create(['class' => 'node-list']));

        $resultSet = $forParents ? $this->parents : $this->children;

        foreach ($resultSet as $dependencyNode) {
            if (isset($dependencyNode->service_id)) {
                $node = $dependencyNode->service;
                $url = Links::service($node, $node->host);
            } elseif (isset($dependencyNode->host_id)) {
                $node = $dependencyNode->host;
                $url = Links::host($node);
            } else {
                $node = $dependencyNode->redundancy_group;
                $url = Url::fromPath('icingadb/redundancygroup', ['id' => bin2hex($node->id)]);
            }

            $list->addHtml(
                new HtmlElement(
                    'li',
                    null,
                    new Link(
                        $this->renderNode($node),
                        $url,
                        Attributes::create(['class' => 'node', 'data-base-target' => '_next'])
                    )
                )
            );
        }

        if ($list->count() > 1) {
            $filterColumn = sprintf(
                '%s.%s.id',
                $forParents ? 'child' : 'parent',
                $this->currentObject->getTableName()
            );

            if ($this->currentObject instanceof RedundancyGroup) {
                $url = Url::fromPath(
                    sprintf('icingadb/redundancygroup/%s', $forParents ? 'members' : 'children'),
                    ['id' => bin2hex($this->currentObject->id)]
                );
            } elseif ($this->currentObject instanceof Service) {
                $url = Url::fromPath(
                    sprintf('icingadb/service/%s', $forParents ? 'parents' : 'children'),
                    ['name' => $this->currentObject->name, 'host.name' => $this->currentObject->host->name]
                );
            } else {
                $url = Url::fromPath(
                    sprintf('icingadb/host/%s', $forParents ? 'parents' : 'children'),
                    ['name' => $this->currentObject->name]
                );
            }

            $summary = DependencyNodeSummary::on($this->getDb())
                ->filter(Filter::equal($filterColumn, $this->currentObject->id));

            $this->applyRestrictions($summary);

            $list->addHtml(
                new HtmlElement(
                    'li',
                    null,
                    new Link(
                        new DependencyNodeStatistics($summary->first()),
                        $url,
                        Attributes::create(['class' => ['summary']])
                    )
                )
            );
        }

        return $list;
    }

    /**
     * Render the given node with link
     *
     * @param Host|Service|RedundancyGroup $node
     *
     * @return ValidHtml
     */
    private function renderNode($node): ValidHtml
    {
        $subject = new HtmlElement('span', Attributes::create(['class' => 'subject']));

        if ($node instanceof Service) {
            $host = [
                new StateBall($node->host->state->getStateText(), StateBall::SIZE_MEDIUM),
                ' ',
                $node->host->display_name
            ];

            $subject
                ->setHtmlContent(Html::sprintf(t('%s on %s', '<service> on <host>'), $node->display_name, $host))
                ->addAttributes([
                    'title' => sprintf(
                        t('%s on %s', '<service> on <host>'),
                        $node->display_name,
                        $node->host->display_name
                    )
                ]);
        } else {
            $subject
                ->setHtmlContent(Text::create(
                    $node instanceof Host
                        ? $node->display_name
                        : sprintf(t('Group: %s'), $node->display_name)
                ))
                ->addAttributes(['title' => $node->display_name]);
        }

        $state = $node->state;
        $mainBall = new StateBall($state->getStateText(), StateBall::SIZE_LARGE);
        $mainBall->setHandled(
            ! $node instanceof RedundancyGroup && $state->is_problem && ($state->is_handled || ! $state->is_reachable)
        );

        if (method_exists($state, 'getIcon')) {
            $mainBall->add($state->getIcon());
        }

        $iconImage = HtmlElement::create('div', ['class' => 'icon-image']);
        if (isset($node->icon_image->icon_image)) {
            $iconImage->addHtml(new IconImage($node->icon_image->icon_image, $node->icon_image_alt));
        }

        return (new HtmlDocument())->addHtml($mainBall, $iconImage, $subject);
    }

    /**
     * Create the SVG lines
     *
     * @param $lineCount int The count of lines to draw
     * @param $forParents bool Whether the svg path should be for parents or children
     *
     * @return HtmlElement
     */
    private function createSVGLines(int $lineCount, bool $forParents = false): HtmlElement
    {
        $path = $forParents
            ? 'M 0 %d C 50 %d 50 50 100 50'
            : 'M 0 50 C 50 50 50 %d 100 %d';

        /**
         * @var $pathCoordinates array<int, array<int>> The `key` contains the count of curves to draw.
         *
         * The Svg Element is a square with fixed width and has a viewBox attr `0 0 100 100`. It is centred vertically
         * to the parent element.
         *
         * When 4 (max.) dependency nodes are present, coordinate 0/100 begins exactly in the middle of
         * the first/last node, coordinate 50 is exactly the center of the parent element. This makes the curve
         * calculation easier.
         */
        $pathCoordinates = [
            0 => [],
            1 => [50],
            2 => [33, 66],
            3 => [17, 50, 83],
            4 => [1, 33, 66, 99]
        ];

        $svg = new HtmlElement('svg', Attributes::create(['viewBox' => '0 0 100 100', 'class' => 'svg-lines']));
        foreach ($pathCoordinates[$lineCount] as $coordinate) {
            $svg->addHtml(
                new HtmlElement(
                    'path',
                    Attributes::create([
                        'stroke-width' => 1,
                        'fill' => 'none',
                        'd' => sprintf($path, $coordinate, $coordinate)
                    ])
                )
            );
        }

        return $svg;
    }
}
