<?php

/* Icinga Dependencies Web | (c) 2024 Icinga GmbH | AGPLv3 */

namespace Icinga\Module\Dependencies\Widget;

use Icinga\Module\Icingadb\Common\Database;
use Icinga\Module\Icingadb\Model\DependencyNode;
use ipl\Html\Attributes;
use ipl\Html\Form;
use ipl\Html\HtmlElement;
use ipl\I18n\Translation;
use ipl\Orm\Exception\InvalidRelationException;
use ipl\Orm\Resolver;
use ipl\Web\Common\FormUid;
use ipl\Web\FormElement\TermInput;
use ipl\Web\Url;
use ipl\Web\Widget\Icon;

class MapSettings extends Form
{
    use Database;
    use FormUid;
    use Translation;

    protected $defaultAttributes = [
        'name' => 'map-settings',
        'class' => ['icinga-controls'],
    ];

    /** @var ?string The group by facts */
    protected $groupBy;

    /**
     * Set the default group by facts
     *
     * @param ?string $groupBy
     *
     * @return $this
     */
    public function setGroupBy(?string $groupBy): self
    {
        $this->groupBy = $groupBy;

        return $this;
    }

    /**
     * Get the default group by facts
     *
     * @return ?string
     */
    public function getGroupBy(): ?string
    {
        return $this->groupBy;
    }

    protected function assemble()
    {
        $this->setDefaultElementDecorator('div');

        $resolver = new Resolver(DependencyNode::on($this->getDb()));

        $groupValidator = function (array $terms) use ($resolver) {
            $knownTerms = [];
            foreach ($terms as $term) {
                /** @var TermInput\ValidatedTerm $term */
                $searchPath = $term->getSearchValue();
                if (strpos($searchPath, '.') === false) {
                    $term->setMessage($this->translate('Is not a valid column'));

                    continue;
                } elseif (isset($knownTerms[$searchPath])) {
                    $term->setMessage($this->translate('You already added this fact'));

                    continue;
                }

                try {
                    $definition = $resolver->getColumnDefinition($searchPath);
                    $term->setLabel($definition->getLabel());
                    $knownTerms[$searchPath] = $term;
                } catch (InvalidRelationException $e) {
                    $term->setMessage($this->translate('Is not a valid column'));
                }
            }

            $existingTerms = array_diff_key($this->getElement('groupBy')->getTerms(), $terms);

            foreach ($existingTerms as $term) {
                if (isset($knownTerms[$term->getSearchValue()])) {
                    $knownTerms[$term->getSearchValue()]->setMessage($this->translate('You already added this fact'));
                }
            }
        };

        $this->addElement(
            (new TermInput('groupBy', [
                'label' => $this->translate('Group By'),
                'value' => $this->getGroupBy()
            ]))
                ->setSuggestionUrl(Url::fromPath('dependencies/map/suggest-group'))
                ->on(TermInput::ON_ENRICH, $groupValidator)
                ->on(TermInput::ON_ADD, $groupValidator)
                ->on(TermInput::ON_SAVE, $groupValidator)
                ->on(TermInput::ON_PASTE, $groupValidator)
        );

        $this->addElement('submit', 'submit', ['label' => $this->translate('Apply')]);
        $this->addElement($this->createUidElement());

        $button = new HtmlElement(
            'button',
            Attributes::create([
                'class' => 'control-button',
                'data-changed-title' => $this->translate('You have unsaved changes')
            ]),
            new HtmlElement('span', Attributes::create(['class' => 'change-indicator'])),
            new Icon('cog')
        );

        $this->prependWrapper(
            (new HtmlElement('div', Attributes::create(['class' => 'map-settings'])))
                ->addHtml($button)
        );
    }

    public function getPartUpdates()
    {
        $this->ensureAssembled();

        foreach ($this->getElement('groupBy')->prepareMultipartUpdate($this->getRequest()) as $update) {
            if (! is_array($update)) {
                $update = [$update];
            }

            yield from $update;
        }
    }
}
