/* Icinga Dependencies Web | (c) 2024 Icinga GmbH | AGPLv3 */
(function (Icinga) {
    define([
            "map/View",
            "map/Utils",
            "icinga/icinga-php-library/iterator",
            "icinga/icinga-php-library/notjQuery",
            "icinga/icinga-php-thirdparty/mbostock/d3"
        ],
        (View, utils, iterator, $, d3) => {

        "use strict";

        const DEBUG = ! document.currentScript.src.endsWith(".min.js");

        class DependenciesMap extends Icinga.EventListener {
            constructor(icinga) {
                super(icinga);

                this._worker = null;
                this._flyout = null;
                this._progress = null;
                this._setupInProgress = false;
                this._stateInProgress = false;
                this._abortController = null;
                this.csrfToken = null;
                this.groupBy = null;
                this.nodeRegistry = new Map();
                this.linkRegistry = new Map();
                this.completionists = new Map();
                this.progressScale = d3.scalePow().range([0, 100]).exponent(2);
                this.view = new View().showWireframe(DEBUG).label(this.fetchNodeLabel.bind(this));
                this.control = null;
                this.baseUrl = location.origin + icinga.config.baseUrl.replace(/\/$/, "");

                this.on("rendered", "#main .container", this.onRendered.bind(this));
                this.on("change", "#lod-control input[type=range]", this.onLodChange.bind(this));
                this.on("click", "#search-results [data-node-id]", this.onNodeShow.bind(this));
                this.on("focusout", "#search-results [data-node-id]", this.onNodeReset.bind(this));
                this.on("click", "#map-container .item-layout:has(a.subject)", this.onNodeLabelClick.bind(this));

                this.configObserver = new MutationObserver(this.onConfigChange.bind(this));
                this.on("click", ".map-settings > button", this.toggleConfigFlyout.bind(this));
                this.on("submit", ".map-settings form", this.hideConfigFlyout.bind(this));
                this.on("mouseenter", ".map-settings", this.showConfigFlyout.bind(this));
                this.on("mouseleave", ".map-settings", this.hideConfigFlyout.bind(this));
                this.on("focusout", ".map-settings", this.hideConfigFlyout.bind(this));
            }

            get worker() {
                if (this._worker === null) {
                    this._worker = new Worker(this.baseUrl + "/dependencies/map/worker");
                }

                return this._worker;
            }

            get flyout() {
                if (this._flyout === null || ! this._flyout.isConnected) {
                    const flyout = document.querySelector("#main > .container > .controls > .map-settings");
                    if (flyout) {
                        this.flyout = flyout;
                    } else {
                        this._flyout = null;
                    }
                }

                return this._flyout;
            }

            set flyout(value) {
                this._flyout = value;

                if (value !== null) {
                    for (const area of value.querySelectorAll(".term-input-area")) {
                        this.configObserver.observe(area, {
                            subtree: true,
                            childList: true,
                            attributeFilter: [ "data-search" ]
                        });
                    }
                }
            }

            get progress() {
                if (this._progress == null) {
                    this._progress = document.getElementById("map-impact");
                } else if (! this._progress.isConnected) {
                    const classes = this._progress.classList.values();
                    this._progress = document.getElementById("map-impact");
                    this._progress.classList.add(...classes);
                }

                return this._progress;
            }

            onRendered(event, autorefresh, scripted, autosubmit) {
                if (event.target !== event.currentTarget) {
                    return;
                }

                const container = event.target.querySelector('#map-container');
                if (container) {
                    if (! autosubmit && this.view.svg.node().isConnected) {
                        // Ensure that it is a different container, not the same that just was not updated…
                        // See https://github.com/Icinga/icingaweb2/issues/5377
                        const otherContainer = this.view.svg.node().closest('.container');
                        if (otherContainer !== event.target) {
                            const content = event.target.querySelector('.content');
                            content.innerHTML = "<div class=\"empty-state-bar\">Only one dependency map at a time is supported</div>";
                            return;
                        }
                    }

                    this.flyout = null;
                    this.view.bindTo(container);
                    this.csrfToken = container.dataset.csrfToken;
                    const invalidate = "invalidate" in container.dataset;
                    const searchResults = JSON.parse(container.dataset.searchResults);

                    this.control = event.target.querySelector('#lod-control input');
                    if (this.control) {
                        this.control.closest("#lod-control").hidden = ! DEBUG;
                    }

                    // Decide whether to update the map or to set it up from scratch
                    const groupBy = container.dataset.groupBy;

                    if (this.view.hasPopulation() && this.groupBy === groupBy) {
                        this.view.update().then(() => {
                            if (! autosubmit && ! autorefresh) {
                                this.applySearchResults(searchResults, true);
                            }
                        });

                        if (this.control) {
                            this.control.max = this.view.countLayers();
                        }

                        if (this._setupInProgress) {
                            this.showProgress();
                        } else {
                            this.fetchStates();
                        }
                    } else {
                        const content = event.target.querySelector('.content');

                        this.setup(groupBy, invalidate)
                            .then(() => {
                                if (searchResults.length) {
                                    this.applySearchResults(searchResults, false);
                                }
                            })
                            .catch(error => {
                                if (error instanceof Error) {
                                    if (error.cause instanceof Error) {
                                        // If cause is an error, it is a PHP exception
                                        container.closest('.container').innerHTML = error.cause.message;
                                    } else if (typeof error.cause !== "undefined") {
                                        // If cause is not an error, it is HTML
                                        content.innerHTML = error.cause;
                                    } else {
                                        // Otherwise, it is a generic error
                                        if (typeof error.stack !== "undefined") {
                                            content.innerText = error.stack;
                                        } else {
                                            content.innerText = error.message;
                                        }

                                        this.icinga.logger.error(error);
                                    }
                                } else {
                                    this.icinga.logger.debug(error);
                                }
                            });
                    }
                }
            }

            isFlyoutOpen() {
                return this.flyout?.classList.contains('open') || false;
            }

            onConfigChange() {
                if (this.flyout !== null && ! this.flyout.classList.contains('changed')) {
                    this.flyout.classList.add('changed');
                    const button = this.flyout.querySelector(":scope > button");
                    button.title = button.dataset.changedTitle;
                }
            }

            toggleConfigFlyout(event) {
                if (this.flyout === null) {
                    return;
                }

                if (this.isFlyoutOpen()) {
                    this.hideConfigFlyout(event);
                } else {
                    this.showConfigFlyout(event);
                }
            }

            showConfigFlyout(event) {
                this.flyout?.classList.add('open');
            }

            hideConfigFlyout(event = null) {
                if (this.flyout !== null) {
                    if (event?.type === "focusout") {
                        setTimeout(() => this.hideConfigFlyout(), 0);

                        return;
                    } else if (event?.type === "submit") {
                        if (event.detail && "submittedBy" in event.detail) {
                            return; // Ignore auto-submit events
                        }

                        this.flyout.classList.remove('changed');
                        this.flyout.querySelector(":scope > button").title = "";
                    } else if (typeof document.activeElement !== "undefined" && this.flyout.contains(document.activeElement)) {
                        return;
                    }

                    this.flyout.classList.remove('open');
                }
            }

            showProgress() {
                if (! this.progress.classList.contains("visible")) {
                    this.progress.classList.add("visible", "indeterminate");
                }
            }

            moveProgress() {
                this.progress.classList.add("moved");
            }

            hideProgress() {
                this.progress.classList.remove("visible", "indeterminate", "moved");
                this.progress.querySelector('.bar').style.width = "";
            }

            updateProgress(tick) {
                this.progress.classList.remove("indeterminate");
                this.progress.querySelector('.bar').style.width = this.progressScale(tick) + "%";
            }

            onLodChange(event) {
                this.view.lod = Number(event.target.value);
                this.view.update();
            }

            onNodeShow(event) {
                const nodeId = event.currentTarget.dataset.nodeId;
                if (this.nodeRegistry.has(nodeId)) {
                    this.view.focusNode(this.nodeRegistry.get(nodeId));
                }

                event.stopPropagation();
                event.stopImmediatePropagation();
            }

            onNodeReset(event) {
                const nodeId = event.currentTarget.dataset.nodeId;
                if (this.nodeRegistry.has(nodeId)) {
                    this.view.blurNode(this.nodeRegistry.get(nodeId));
                }
            }

            onNodeLabelClick(event) {
                const anchor = event.currentTarget.querySelector("a.subject");
                if (anchor) {
                    anchor.click();
                }
            }

            setup(groupBy, invalidate) {
                if (this._setupInProgress) {
                    if (this.groupBy === groupBy) {
                        this.showProgress();

                        return Promise.reject("Setup already in progress");
                    }

                    this.abortSetup();
                }

                this._setupInProgress = true;
                this._abortController = new AbortController();
                this.groupBy = groupBy;

                const abortSelf = new AbortController();
                this._abortController.signal.addEventListener("abort", () => abortSelf.abort());

                return new Promise((resolve, reject) => {
                    this.worker.postMessage({
                        baseUrl: this.baseUrl,
                        csrfToken: encodeURIComponent(this.csrfToken),
                        group: this.groupBy ? encodeURIComponent(this.groupBy) : null,
                        invalidate,
                        debug: DEBUG
                    });

                    if (invalidate) {
                        // Reset the view to avoid stale data
                        this.view.reset(false);
                    }

                    this.showProgress();

                    this.worker.addEventListener("message", event => {
                        if (! ("data" in event) || typeof event.data !== "object") {
                            this.icinga.logger.debug("Unexpected message received from worker", event);
                        } else if ("error" in event.data) {
                            this.icinga.logger.error(event.data.error);
                            reject(event.data.error);
                        } else if ("networkIdle" in event.data) {
                            // Fetch in parallel, do not block the UI
                            this.fetchStates(false);
                        } else if ("totalTicks" in event.data) {
                            this.progressScale.domain([1, event.data.totalTicks]);
                        } else if ("tick" in event.data) {
                            this.updateProgress(event.data.tick);
                        } else if ("success" in event.data) {
                            abortSelf.abort();

                            resolve();

                            if (! event.data.success) {
                                this.icinga.loader.createNotice("error", event.data.reason, true);
                            }
                        } else {
                            const { layer, nodes, links } = event.data;

                            for (let i = nodes.length - 1; i >= 0; i--) {
                                const node = nodes[i];
                                if (this.nodeRegistry.has(node.id)) {
                                    // References in the registry must be kept intact at all times. If such exist
                                    // already, the data returned by the worker must be merged into the existing
                                    // object. Lastly, the array entry must be overridden with the updated object.
                                    nodes[i] = Object.assign(this.nodeRegistry.get(node.id), node);

                                    this.complete(node.id);
                                } else {
                                    this.nodeRegistry.set(node.id, node);
                                }
                            }

                            for (let i = links.length - 1; i >= 0; i--) {
                                const link = links[i];
                                if (this.linkRegistry.has(link.stateId)) {
                                    const data = this.linkRegistry.get(link.stateId).get("data");
                                    if ("state" in data) {
                                        // State got registered before all links were
                                        // This also means, total_children is done
                                        link.failed = data.state;
                                    } else {
                                        data.total_children = (data.total_children ?? 0) + 1;
                                    }

                                    this.linkRegistry.get(link.stateId).set(link.id, link);

                                    this.complete(link.stateId);
                                } else {
                                    this.linkRegistry.set(link.stateId, new Map([
                                        [ "data", { total_children: 1 } ],
                                        [ link.id, link ]
                                    ]));
                                }

                                if (links[i]?.blob) {
                                    // This gets the job done at the moment, but once cross-layer links are a thing,
                                    // the counterpart (i.e. a link is registered before the node) must be handled.
                                    this.nodeRegistry.get(link.source_id).link = links[i];
                                    delete links[i].blob;
                                }
                            }

                            if (nodes.length > 0) {
                                if (this.control) {
                                    this.control.max = Number(this.control.max || 0) + 1;
                                }

                                const p = this.view.populate(layer, nodes, links);
                                if (this.view.countLayers() === 1) {
                                    if (invalidate) {
                                        p.then(() => this.view.fit(nodes));
                                    } else {
                                        // Do not animate fit upon initial setup
                                        this.view.fit(nodes, false);
                                    }
                                }
                            } else {
                                this.view.clearLayer(layer);

                                if (this.control) {
                                    this.control.max = Number(this.control.max || 1) - 1;
                                }
                            }

                            this.moveProgress();
                        }
                    }, { signal: abortSelf.signal });
                }).finally(() => {
                    this._setupInProgress = false;
                    this._abortController = null;
                    this.hideProgress();
                });
            }

            fetchStates(allowInvalidation = true) {
                if (allowInvalidation && this._setupInProgress) {
                    return Promise.reject();
                }

                this._stateInProgress = true;

                const _this = this;
                return fetch(this.baseUrl + "/dependencies/map/states", { signal: this._abortController?.signal })
                    .then(async response => {
                        if (response.status !== 200) {
                            _this.icinga.logger.error(await response.text());
                            _this.icinga.loader.createNotice(
                                "error", "Failed to fetch dependency map states: " + response.statusText, true);
                        } else {
                            // If any of the sets is not empty, it means a node/link has been removed in the backend
                            const remainingLinks = new Set(allowInvalidation ? _this.linkRegistry.keys() : []);
                            const remainingNodes = new Set();
                            const remainingCluster = new Set();
                            for (const node of _this.nodeRegistry.values()) {
                                if (node.type === "blob") {
                                    remainingCluster.add(node.id);
                                } else if ("state" in node) {
                                    remainingNodes.add(node.id);
                                }
                            }

                            let invalidateStructure = false; // true in case the backend transmits a new node/link
                            for await (const line of utils.readLines(response.body)) {
                                const data = JSON.parse(line);
                                if (data[0] === 1) { // node state
                                    const [ _, id, type, state ] = data;
                                    if (_this.nodeRegistry.has(id)) {
                                        const node = _this.nodeRegistry.get(id);
                                        Object.assign(node, { type, state });

                                        remainingNodes.delete(id);
                                        this.complete(id);
                                    } else {
                                        _this.nodeRegistry.set(id, { id, type, state });

                                        if (allowInvalidation) {
                                            invalidateStructure = true;
                                        }
                                    }
                                } else if (data[0] === 2) { // edge state
                                    const [ _, stateId, state, total_children = 0 ] = data;
                                    if (_this.linkRegistry.has(stateId)) {
                                        const links = _this.linkRegistry.get(stateId);
                                        links.get("data").state = state;
                                        if (total_children !== links.get("data").total_children) {
                                            if (allowInvalidation) {
                                                invalidateStructure = true;
                                            } else {
                                                // If invalidation is not allowed, it means structure is not final yet.
                                                // By registering the state in "data" now, setup() can stop counting
                                                // children since it is already done by the backend here.
                                                links.get("data").total_children = total_children;
                                            }
                                        }

                                        // Update already registered links
                                        for (const [ linkId, link ] of links) {
                                            if (linkId !== "data") {
                                                link.failed = state;
                                            }
                                        }

                                        remainingLinks.delete(stateId);
                                        this.complete(stateId);
                                    } else {
                                        _this.linkRegistry.set(stateId, new Map(
                                            [ [ "data", { state, total_children } ] ]
                                        ));

                                        if (allowInvalidation) {
                                            invalidateStructure = true;
                                        }
                                    }
                                }
                            }

                            if (invalidateStructure || remainingNodes.size > 0 || remainingLinks.size > 0) {
                                // Delete nodes/links that were not in the response, but keep others to not lose state
                                for (const nodeId of remainingNodes) {
                                    _this.nodeRegistry.delete(nodeId);
                                }

                                for (const stateId of remainingLinks) {
                                    _this.linkRegistry.delete(stateId);
                                }

                                // Cluster can be deleted without checking whether any member remains.
                                // A structure invalidation causes them to be re-created anyway.
                                for (const nodeId of remainingCluster) {
                                    _this.nodeRegistry.delete(nodeId);
                                }

                                return _this.setup(this.groupBy, true).then(() => {
                                    const container = document.getElementById("map-container");
                                    const searchResults = JSON.parse(container.dataset.searchResults);
                                    if (searchResults.length) {
                                        _this.applySearchResults(searchResults, true);
                                    }
                                });
                            } else {
                                return _this.view.update();
                            }
                        }
                    }).catch(error => {
                        if (error.name !== "AbortError") {
                            throw error;
                        }
                    }).finally(() => {
                        _this._stateInProgress = false;
                    });
            }

            /**
             * Wait for the setup and state to complete a particular node
             * @param {*} id The id of the node
             * @returns {Promise<void>}
             */
            waitForCompletion(id) {
                if (! this._setupInProgress && ! this._stateInProgress) {
                    return Promise.resolve();
                }

                if (this._stateInProgress && this.nodeRegistry.has(id) && "state" in this.nodeRegistry.get(id)) {
                    // Waiting for state to be updated is not required, as usually having state is a sign
                    // of the user to be authorized to see the node. So a node can be considered complete,
                    // even if the state is not up to date.
                    return Promise.resolve();
                }

                if (this.completionists.has(id)) {
                    return this.completionists.get(id)[0];
                }

                const completionist = {};
                completionist.promise = new Promise(resolve => {
                    completionist.resolve = resolve;
                });

                this.completionists.set(id, completionist);

                return completionist.promise;
            }

            /**
             * Mark a node as complete
             * @param {*} id The id of the node
             */
            complete(id) {
                if (this.completionists.has(id)) {
                    this.completionists.get(id).resolve();
                    this.completionists.delete(id);
                }
            }

            abortSetup() {
                if (this._abortController !== null) {
                    this._abortController.abort();
                    this._abortController = null;
                }

                this.worker.terminate();
                this._worker = null;
                this.hideProgress();
                this.view.reset(false);
            }

            async fetchNodeLabel(data) {
                await this.waitForCompletion(data.id);

                if (! ['host', 'service', 'redundancy_group'].includes(data?.type)) {
                    return Promise.resolve(null); // Only supported nodes
                } else if (! ("state" in data)) {
                    return Promise.resolve(null); // Only authorized nodes
                }

                const url = this.baseUrl + `/dependencies/map/node-label?id=${data.id}`;

                return fetch(url).then(async response => $.render(await response.text()));
            }

            /**
             * Propagate search results, provided by the backend as part of a multipart update, to the view
             * @param {Array<*>} ids
             */
            update(ids) {
                this.applySearchResults(ids, true);
            }

            /**
             * Apply search results to the view
             * @param {Array<*>} results IDs of nodes that match the search query
             * @param {boolean} animate Whether to animate the transition
             */
            applySearchResults(results, animate) {
                if (! results.length) {
                    this.view.highlight([]); // Must happen before fit
                    this.view.fit([], animate);
                    this.view.blurNode();

                    return;
                }

                const ids = [];
                const nodes = [];
                for (const id of results) {
                    if (this.nodeRegistry.has(id)) {
                        ids.push(id);
                        nodes.push(this.nodeRegistry.get(id));
                    }
                }

                this.view.highlight(ids);

                if (nodes.length > 1) {
                    this.view.fit(nodes, animate);
                } else if (nodes.length) {
                    this.view.focusNode(nodes[0], animate);
                }
            }

            unbind(emitter) {
                super.unbind(emitter);

                this.view.reset();
                this.nodeRegistry.clear();
                this.linkRegistry.clear();
                this.configObserver.disconnect();
                this.worker.terminate();
            }
        }

        Icinga.Behaviors = Icinga.Behaviors || {};

        Icinga.Behaviors.DependenciesMap = DependenciesMap;

        return DependenciesMap;
    });
})(Icinga);
