/* Icinga Dependencies Web | (c) 2024 Icinga GmbH | AGPLv3 */
(function (root, factory) {
    if (typeof define === "function" && define.icinga) {
        define([
            "icinga/icinga-php-thirdparty/mbostock/d3",
            "Utils",
            "icinga/icinga-php-library/iterator"
        ], factory);
    } else {
        root.icingaDependencyMapGroupForce = factory(
            root.d3 = root.d3 || {},
            root.icingaDependencyMapUtils = root.icingaDependencyMapUtils || {},
            root.icingaIteratorPolyfill = root.icingaIteratorPolyfill || {}
        );
    }
}(self, function (d3, utils, iterator) {
    "use strict";

    /**
     * Create and return a new group force.
     * @param {object[]} _links
     * @returns {force}
     */
    function factory(_links) {
        let nodes,
            groupSimulation,
            groups = [],
            hierarchies = [],
            links = _links,
            radius = d => 1,
            totalTicks = 300,
            visualizeAnchors = false,
            groupPadding = 100,
            alphaThreshold = 0.75,
            alphaMin = 0.001,
            alphaTarget = 0,
            alphaDecay = 1 - Math.pow(alphaMin, 1 / 300),
            linkIterations = 1,
            linkDistance = 30,
            linkStrength,
            hierarchyIterations = 2,
            hierarchyStrength = 1,
            hierarchyPadding = 10;

        function remainingTicks(alpha) {
            return Math.ceil(Math.log((alphaMin - alphaTarget) / (alpha - alphaTarget)) / Math.log(1 - alphaDecay));
        }

        const force = alpha => {
            const ticksRemaining = remainingTicks(alpha);
            const alphaProgress = 1 - ticksRemaining / totalTicks;
            const alphaThresholdCrossed = alphaProgress >= alphaThreshold;

            groupSimulation.alpha(alpha).tick();

            if (! alphaThresholdCrossed) {
                for (const [ memberSimulation, _ ] of hierarchies) {
                    memberSimulation.alpha(alpha).tick();
                }
            } else {
                if (alphaProgress === alphaThreshold) {
                    force.finalizeHierarchies();
                }

                for (const [ hierarchySimulation, groupAnchor ] of groups) {
                    const hierarchyNodes = hierarchySimulation.nodes();
                    const occupiedArea = utils.occupiedAnchorArea(groupAnchor, hierarchyNodes, hierarchyPadding);
                    hierarchySimulation.force("manyBody").strength(occupiedArea < 0.8 ? 10 : 0);
                    hierarchySimulation.force("center").x(groupAnchor.x).y(groupAnchor.y);
                    hierarchySimulation.alpha(alpha).tick();

                    const misfits = [];
                    for (const node of hierarchyNodes) {
                        const d = utils.distanceBetween(node, groupAnchor) + node.r;
                        if (d > groupAnchor.r) {
                            misfits.push(node);
                        }
                    }

                    // Enlarge the anchor if there are too many misfits
                    if (misfits.length / hierarchyNodes.length > 0.1 && occupiedArea >= 0.8
                        || ! misfits.length && occupiedArea < 0.8 // or shrink it if the many body force does its work
                        || ticksRemaining === 10 // and just fit it if the simulation is almost done
                    ) {
                        const coveringCircle = d3.packEnclose(hierarchyNodes);
                        groupAnchor.r = coveringCircle.r;
                        groupAnchor.vx += groupAnchor.x - coveringCircle.x;
                        groupAnchor.vy += groupAnchor.y - coveringCircle.y;
                    }
                }

                for (const [ memberSimulation, hierarchyAnchor ] of hierarchies) {
                    // In order to keep anchors close together, we need to reset the velocities ourselves.
                    // d3's velocity decay is applied too early so if we set it to 1, anchors don't move at all.
                    hierarchyAnchor.vx = 0;
                    hierarchyAnchor.vy = 0;

                    // Move nodes with their respective anchors
                    for (const node of memberSimulation.nodes()) {
                        node.x = hierarchyAnchor.x + node.xOffset;
                        node.y = hierarchyAnchor.y + node.yOffset;
                    }
                }
            }

            // The group anchor radii might have changed
            groupSimulation.force("collide").radius(d => d.r + groupPadding);
        }

        force.initialize = _nodes => {
            nodes = _nodes;

            const nodeGroups = utils.createGroups(nodes, links);

            const crossGroupLinks = [];
            const groupHierarchies = new Map(iterator.map(nodeGroups.keys(), name => [ name, [] ]));
            for (const [ ids, memberLinks ] of utils.isolateHierarchies(links)) {
                let members = new Set(ids);
                let links = new Set(memberLinks);

                for (const [ name, groupSet ] of nodeGroups) {
                    if (groupSet.isSupersetOf(members)) {
                        const enclosedLinks = iterator.filter(
                            links.values(), l => members.has(l.source_id) && members.has(l.target_id));
                        const crossLinks = iterator.filter(
                            links.values(), l => ! members.has(l.source_id) || ! members.has(l.target_id));

                        crossGroupLinks.push(...crossLinks);
                        groupHierarchies.get(name).push([ [ ...members ], [ ...enclosedLinks ] ]);

                        members.clear();

                        break;
                    } else if (groupSet.isSubsetOf(members)) {
                        members = members.difference(groupSet);
                        const otherLinks = new Set(iterator.filter(
                            links.values(), l => groupSet.has(l.source_id) && groupSet.has(l.target_id)));
                        links = links.difference(otherLinks);

                        groupHierarchies.get(name).push([ [ ...groupSet ], [ ...otherLinks ] ]);
                    } else {
                        const intersection = groupSet.intersection(members);
                        if (intersection.size) {
                            const enclosedLinks = new Set(iterator.filter(
                                links.values(), l => intersection.has(l.source_id) && intersection.has(l.target_id)));

                            groupHierarchies.get(name).push([ [ ...intersection ], [ ...enclosedLinks ] ]);

                            members = members.difference(intersection);
                            links = links.difference(enclosedLinks);
                        }
                    }
                }
            }

            let i = -1;
            const idToNodeMap = new Map(_nodes.map(d => [ d.id, d ]));
            for (const [ name, groupMembers ] of groupHierarchies) {
                const hierarchyAnchors = [];
                for (const [ members, memberLinks ] of groupMembers) {
                    const memberNodes = members.map(id => idToNodeMap.get(id));
                    const memberSimulation = d3.forceSimulation(memberNodes).stop();
                    memberSimulation.force("link", d3.forceLink(memberLinks)
                        .id(d => d.id)
                        .iterations(linkIterations)
                        .distance(l => radius(l.source) + radius(l.target) + linkDistance));
                    if (linkStrength) {
                        memberSimulation.force("link").strength(linkStrength);
                    }

                    memberSimulation.force("collide", d3.forceCollide()
                        .radius(d => radius(d) + (linkDistance / 2))
                        .strength(hierarchyStrength)
                        .iterations(hierarchyIterations));

                    const anchor = { r: 11, id: i-- };
                    hierarchies.push([ memberSimulation, anchor ]);
                    hierarchyAnchors.push(anchor);
                }

                const hierarchySimulation = d3.forceSimulation(hierarchyAnchors)
                    .velocityDecay(0.1) // Low friction, so the anchors can move freely
                    .stop();

                groups.push([
                    hierarchySimulation,
                    { ...d3.packEnclose(hierarchyAnchors), id: name }
                ]);
            }

            groupSimulation = d3.forceSimulation(groups.map(g => g[1]))
                .force("collide", d3.forceCollide().radius(d => d.r + groupPadding))
                .force("center", d3.forceCenter()) // Keep groups around 0,0
                .stop();

            if (crossGroupLinks.length) {
                const groupLinks = [];
                const linkedGroups = new Map();
                for (const groupLink of crossGroupLinks) {
                    const source = idToNodeMap.get(groupLink.source_id);
                    const target = idToNodeMap.get(groupLink.target_id);
                    groupLink.source = source; // There's no link force involved this time,
                    groupLink.target = target; // so we have to do it ourselves

                    const sourceGroup = iterator.find(nodeGroups.entries(), e => e[1].has(source.id));
                    const targetGroup = iterator.find(nodeGroups.entries(), e => e[1].has(target.id));

                    // No duplicates please
                    if (! linkedGroups.has(sourceGroup[0]) || ! linkedGroups.get(sourceGroup[0]).has(targetGroup[0])) {
                        if (! linkedGroups.has(sourceGroup[0])) {
                            linkedGroups.set(sourceGroup[0], new Set());
                        }
                        linkedGroups.get(sourceGroup[0]).add(targetGroup[0]);
                    } else if (! linkedGroups.has(targetGroup[0]) || ! linkedGroups.get(targetGroup[0]).has(sourceGroup[0])) {
                        if (! linkedGroups.has(targetGroup[0])) {
                            linkedGroups.set(targetGroup[0], new Set());
                        }
                        linkedGroups.get(targetGroup[0]).add(sourceGroup[0]);
                    } else {
                        continue;
                    }

                    groupLinks.push({ source: sourceGroup[0], target: targetGroup[0] });
                }

                groupSimulation.force("link", d3.forceLink(groupLinks)
                    .id(d => d.id)
                    .distance(d => d.source.r + d.target.r + linkDistance));
            }
        }

        force.finalizeHierarchies = () => {
            for (const [ memberSimulation, hierarchyAnchor ] of hierarchies) {
                const coveringCircle = d3.packEnclose(
                    memberSimulation.nodes().map(d => ({ x: d.x, y: d.y, r: radius(d) })));
                hierarchyAnchor.r = coveringCircle.r;

                for (const node of memberSimulation.nodes()) {
                    // Measure the initial offset of each node to its anchor,
                    // in order to keep them together wherever the anchor is moving
                    node.xOffset = node.x - coveringCircle.x;
                    node.yOffset = node.y - coveringCircle.y;

                    // Initially force the nodes to the center of the covering circle
                    node.x = hierarchyAnchor.x + node.xOffset;
                    node.y = hierarchyAnchor.y + node.yOffset;
                }
            }

            for (const [ hierarchySimulation, _ ] of groups) {
                // Now that the hierarchy members are laid out and synced with their anchor position, keep them there
                hierarchySimulation.force("center", d3.forceCenter().strength(1));
                // And since the member anchors have a proper radius, make sure they don't overlap
                hierarchySimulation.force("collide", d3.forceCollide().iterations(4).radius(d => d.r + hierarchyPadding));
                // But this may cause the anchors to be pushed too far apart, so we need to counteract that, if necessary
                hierarchySimulation.force("manyBody", d3.forceManyBody().strength(0).distanceMin(
                    d3.median(hierarchySimulation.nodes(), d => d.r + hierarchyPadding) * 2 + hierarchyPadding
                ));
            }
        }

        /**
         * Set the radius function for the nodes.
         * @param r
         * @returns {force}
         */
        force.radius = r => {
            radius = typeof r === "function" ? r : () => Number(r);

            return force;
        }

        /**
         * Set the amount of padding between groups.
         * @param p
         * @returns {force}
         */
        force.groupPadding = p => {
            groupPadding = Number(p);

            return force;
        }

        /**
         * Set the threshold for the alpha value at which the hierarchy layout is started.
         * @param a
         * @returns {force}
         */
        force.alphaThreshold = a => {
            alphaThreshold = Number(a);

            return force;
        }

        /**
         * Set the minimum alpha value.
         * @param m
         * @returns {force}
         */
        force.alphaMin = m => {
            alphaMin = m;
            totalTicks = remainingTicks(1);

            return force;
        }

        /**
         * Set the target alpha value.
         * @param t
         * @returns {force}
         */
        force.alphaTarget = t => {
            alphaTarget = t;
            totalTicks = remainingTicks(1);

            return force;
        }

        /**
         * Set the alpha decay value.
         * @param d
         * @returns {force}
         */
        force.alphaDecay = d => {
            alphaDecay = d;
            totalTicks = remainingTicks(1);

            return force;
        }

        /**
         * Set whether to visualize the anchors.
         * @param s
         * @returns {force}
         */
        force.visualizeAnchors = s => {
            visualizeAnchors = s;

            return force;
        }

        /**
         * Set the number of iterations for a hierarchy's link force.
         * @param i
         * @returns {force}
         */
        force.linkIterations = i => {
            linkIterations = Number(i);

            return force;
        }

        /**
         * Set the distance for a hierarchy's link force.
         * @param d
         * @returns {force}
         */
        force.linkDistance = d => {
            linkDistance = Number(d);

            return force;
        }

        /**
         * Set the strength for a hierarchy's link force.
         * @param s
         * @returns {force}
         */
        force.linkStrength = s => {
            linkStrength = Number(s);

            return force;
        }

        /**
         * Set the number of iterations for a hierarchy's collide force.
         * @param i
         * @returns {force}
         */
        force.hierarchyIterations = i => {
            hierarchyIterations = Number(i);

            return force;
        }

        /**
         * Set the strength for a hierarchy's collide force.
         * @param s
         * @returns {force}
         */
        force.hierarchyStrength = s => {
            hierarchyStrength = Number(s);

            return force;
        }

        /**
         * Set the amount of padding between hierarchy members.
         * @param p
         * @returns {force}
         */
        force.hierarchyPadding = p => {
            hierarchyPadding = Number(p);

            return force;
        }

        /**
         * Get anchors used by this force.
         * @returns {object[]}
         */
        force.anchors = () => {
            if (! visualizeAnchors) {
                return [];
            }

            const anchorNodes = [];
            for (let [ _, groupAnchor ] of groups) {
                anchorNodes.push(groupAnchor);
            }

            for (let [ _, hierarchyAnchor ] of hierarchies) {
                anchorNodes.push(hierarchyAnchor);
            }

            return anchorNodes;
        }

        force.groups = () => {
            if (groups.length === 1) {
                return [];
            }

            const groupAnchors = [];
            for (const [ _, anchor ] of groups) {
                const group = { ...anchor, type: "group" };
                delete group.nodes;
                groupAnchors.push(group);
            }

            return groupAnchors;
        }

        return force;
    }

    return factory;
}));
