/* Icinga Dependencies Web | (c) 2024 Icinga GmbH | AGPLv3 */
(function (root, factory) {
    if (typeof define === "function" && define.icinga) {
        define([
            "exports",
            "icinga/icinga-php-library/iterator",
            "icinga/icinga-php-thirdparty/mbostock/d3"
        ], factory);
    } else {
        factory(root.icingaDependencyMapUtils = root.icingaDependencyMapUtils || {},
            root.icingaIteratorPolyfill = root.icingaIteratorPolyfill || {},
            root.d3);
    }
}(self, function (exports, iterator, d3) {
    "use strict";

    /**
     * Group the given nodes but preserve links.
     * @param {object[]} nodes
     * @param {object[]} links
     * @returns {Map<string, Number[]>} Group names as keys and node ids as values.
     */
    function createGroups(nodes, links) {
        const idToNodeMap = new Map(iterator.map(nodes, d => [ d.id, d ]));
        const groups = new Map();

        for (const link of links) {
            let group = Array.from((new Set([
                ...idToNodeMap.get(link.source_id).group,
                ...idToNodeMap.get(link.target_id).group
            ]))).sort().join("|");

            const sourceGroup = iterator.find(groups.entries(), e => e[1].has(link.source_id));
            const targetGroup = iterator.find(groups.entries(), e => e[1].has(link.target_id));

            if (sourceGroup !== undefined) {
                if (targetGroup === undefined) {
                    groups.get(sourceGroup[0]).add(link.target_id);
                }
            } else if (targetGroup !== undefined) {
                groups.get(targetGroup[0]).add(link.source_id);
            } else {
                if (! groups.has(group)) {
                    groups.set(group, new Set());
                }

                groups.get(group).add(link.source_id);
                groups.get(group).add(link.target_id);
            }
        }

        return groups;
    }

    /**
     * Isolate hierarchies from a set of links.
     * @param {object[]} links
     * @returns {Array<Array<Number,Number>,object[]>}
     */
    function isolateHierarchies(links) {
        const map = new Map();
        const hierarchies = [];
        for (const link of links) {
            const source = map.get(link.source_id);
            const target = map.get(link.target_id);
            if (source !== undefined && target !== undefined) {
                if (source !== target) {
                    hierarchies[source][0].push(...hierarchies[target][0]);
                    hierarchies[source][1].push(link, ...hierarchies[target][1]);
                    hierarchies[target][0].forEach(id => map.set(id, source));
                    hierarchies[target] = null;
                } else if (! hierarchies[source][1].includes(link)) {
                    hierarchies[source][1].push(link);
                }
            } else if (source !== undefined) {
                hierarchies[source][0].push(link.target_id);
                hierarchies[source][1].push(link);
                map.set(link.target_id, source);
            } else if (target !== undefined) {
                hierarchies[target][0].push(link.source_id);
                hierarchies[target][1].push(link);
                map.set(link.source_id, target);
            } else {
                map.set(link.source_id, hierarchies.length);
                map.set(link.target_id, hierarchies.length);
                hierarchies.push([ [ link.source_id, link.target_id ], [ link ] ]);
            }
        }

        return hierarchies.filter(h => h !== null);
    }

    /**
     * Calculate the distance between two points.
     * @param {object} a
     * @param {object} b
     * @returns {Number}
     */
    function distanceBetween(a, b) {
        const dx = a.x - b.x;
        const dy = a.y - b.y;

        return Math.sqrt(dx * dx + dy * dy);
    }

    /**
     * Calculate the circumference point of the target node relative to the source node.
     * @param {Number} x1 The x-coordinate of the target node
     * @param {Number} y1 The y-coordinate of the target node
     * @param {Number} x2 The x-coordinate of the source node
     * @param {Number} y2 The y-coordinate of the source node
     * @param {Number} rx The radius of the target node
     * @returns {Array<Number>}
     */
    function circumferencePoint(x1, y1, x2, y2, rx) {
        const dx = x1 - x2;
        const dy = y1 - y2;
        const gamma = Math.atan2(dy, dx);

        return [ x1 - (Math.cos(gamma) * rx), y1 - (Math.sin(gamma) * rx) ];
    }

    /**
     * Calculate four points that are perpendicular on both ends to a line between the target and source node.
     * @param {{x: Number, y: Number, r: Number}} target The target node
     * @param {{x: Number, y: Number, r: Number}} source The source node
     * @param {Number} targetDistance The desired distance between the two points on the target node
     * @param {Number} sourceDistance The desired distance between the two points on the source node
     * @returns {array<{x: Number, y: Number}>}
     */
    function perpendicularPoints(target, source, targetDistance, sourceDistance) {
        const [ sx, sy ] = circumferencePoint(source.x, source.y, target.x, target.y, source.r);
        const [ tx, ty ] = circumferencePoint(target.x, target.y, source.x, source.y, target.r);

        const points = [];
        for (const [ x1, y1, x2, y2, distance ] of [
            [ tx, ty, sx, sy, targetDistance ],
            [ sx, sy, tx, ty, sourceDistance ]
        ]) {

            const theta = Math.atan2(y2 - y1, x2 - x1);
            const dx = Math.cos(theta + Math.PI / 2);
            const dy = Math.sin(theta + Math.PI / 2);

            points.push(
                {
                    x: roundNumber(x1 + (distance / 2) * dx, 2),
                    y: roundNumber(y1 + (distance / 2) * dy, 2)
                },
                {
                    x: roundNumber(x1 - (distance / 2) * dx, 2),
                    y: roundNumber(y1 - (distance / 2) * dy, 2)
                }
            );
        }

        return points;
    }

    /**
     * Calculate the occupied area of an anchor by a set of hierarchies.
     * @param {object} anchor
     * @param {object[]} hierarchies
     * @param {Number} padding The padding between hierarchies.
     * @returns {Number} The occupied area as a fraction of the anchor area.
     */
    function occupiedAnchorArea(anchor, hierarchies, padding) {
        const anchorArea = Math.PI * anchor.r * anchor.r;
        let occupiedArea = 0;
        for (const node of hierarchies) {
            occupiedArea += Math.PI * ((node.r + padding) ** 2);
        }

        return clamp(occupiedArea / anchorArea, 0, 1);
    }

    /**
     * Shuffle a string.
     * @param {string} str
     * @returns {string}
     */
    function shuffleString(str) {
        const a = str.split(""),
            n = a.length;

        for (let i = n - 1; i > 0; i--) {
            const j = Math.floor(Math.random() * (i + 1));
            const tmp = a[i];
            a[i] = a[j];
            a[j] = tmp;
        }

        return a.join("");
    }

    /**
     * Generate a random id of the given length.
     * @param {Number} len
     * @returns {string}
     */
    function generateId(len) {
        return shuffleString('abcefghijklmnopqrstuvwxyz').substring(0, len);
    }

    /**
     * Read chunks from the given stream
     * @param {ReadableStream} stream
     * @returns {AsyncGenerator<string, void, *>}
     */
    async function* readStream(stream) {
        const reader = stream.pipeThrough(new TextDecoderStream()).getReader();

        while (true) {
            const { value: chunk, done: noMoreData } = await reader.read();
            if (noMoreData) {
                break;
            } else {
                yield chunk;
            }
        }
    }

    /**
     * Read chunks separated by line feeds (aka lines) from the given stream
     * @param {ReadableStream} stream
     * @returns {AsyncGenerator<string, void, *>}
     */
    async function* readLines(stream) {
        const newline = /\n/gm;

        let buffer = "";
        for await (const chunk of readStream(stream)) {
            buffer += chunk;

            let startIndex = 0;
            while (newline.test(buffer)) {
                yield buffer.substring(startIndex, newline.lastIndex);
                startIndex = newline.lastIndex;
            }

            if (startIndex > 0) {
                buffer = buffer.substring(startIndex);
            }
        }

        // If, for some reason, the stream does not end with a newline, yield the remaining buffer
        if (buffer.length > 0) {
            yield buffer;
        }
    }

    /**
     * Find nodes that are visible in the given view frame
     * @param {d3.quadtree} quadtree
     * @param {d3.ZoomTransform} transform
     * @param {Number} margin Enlarge the view frame by this factor
     * @param {Number} x
     * @param {Number} y
     * @param {Number} width
     * @param {Number} height
     * @returns {Set<any>}
     */
    function findVisibleNodes(quadtree, transform, margin, x, y, width, height) {
        const xMin = transform.invertX(x);
        const xMax = transform.invertX(x + width);
        const yMin = transform.invertY(y);
        const yMax = transform.invertY(y + height);
        const xBuffer = width * margin;
        const yBuffer = height * margin;

        const visibleNodes = new Set();
        quadtree.visit((node, x1, y1, x2, y2) => {
            if (! node.length) {
                do {
                    if (node.data.x >= xMin && node.data.x < xMax && node.data.y >= yMin && node.data.y < yMax) {
                        // The center of the node is within the view
                        visibleNodes.add(node.data.id);
                    } else {
                        // A circle's boundary might still be within the view
                        if ((node.data.x < xMin && node.data.x + node.data.r >= xMin)
                            || (node.data.x >= xMax && node.data.x - node.data.r < xMax)
                            || (node.data.y < yMin && node.data.y + node.data.r >= yMin)
                            || (node.data.y >= yMax && node.data.y - node.data.r < yMax)
                        ) {
                            visibleNodes.add(node.data.id);
                        }
                    }
                } while ((node = node.next));

                return false;
            }

            if (x1 > xMax || x2 < xMin || y1 > yMax || y2 < yMin) {
                return true; // The node is fully outside the view
            } else if (x1 < xMin && x2 > xMax || y1 < yMin && y2 > yMax) {
                return false; // The node is larger than the view (e.g. the root)
            }

            // Only visit nodes that are at least partially visible to a given threshold
            return xMax - x1 < xBuffer || yMax - y1 < yBuffer || x2 - xMin < xBuffer || y2 - yMin < yBuffer;
        });

        return visibleNodes;
    }

    /**
     * Round a number to a given precision.
     * From https://stackoverflow.com/a/48764436
     * @param {Number} number
     * @param {Number} precision
     * @returns {Number}
     */
    function roundNumber(number, precision = 0) {
        const p = Math.pow(10, precision);
        const n = (number * p) * (1 + Number.EPSILON);
        return Math.round(n) / p;
    }

    /**
     * Clamp a value to a given range.
     * @param {Number} value
     * @param {Number} min
     * @param {Number} max
     * @returns {Number}
     */
    function clamp(value, min, max) {
        return Math.max(min, Math.min(max, value));
    }

    exports.clamp = clamp;
    exports.roundNumber = roundNumber;
    exports.createGroups = createGroups;
    exports.isolateHierarchies = isolateHierarchies;
    exports.distanceBetween = distanceBetween;
    exports.circumferencePoint = circumferencePoint;
    exports.perpendicularPoints = perpendicularPoints;
    exports.occupiedAnchorArea = occupiedAnchorArea;
    exports.generateId = generateId;
    exports.readStream = readStream;
    exports.readLines = readLines;
    exports.findVisibleNodes = findVisibleNodes;
}));
