/* Icinga Dependencies Web | (c) 2024 Icinga GmbH | AGPLv3 */
define([
    "icinga/icinga-php-thirdparty/mbostock/d3",
    "icinga/icinga-php-library/iterator",
    "Utils"
], (d3, iterator, utils) => {
    "use strict";

    /**
     * The wireframe's cell width and height
     * @type {number}
     */
    const WIREFRAME_CELL_SIZE = 50;

    class View {
        constructor() {
            this.nodeBorderWidth = 1;
            this._box = [-960, -540, 1920, 1080];

            // TODO: Decide which to eventually keep, if any
            this._blobDonuts = true;
            this._blobSummaries = false;

            this._showWireframe = false;
            this._lod = null;
            this._lodLevels = new Map();
            this._highlightPermanently = new Set();
            this._highlightTemporary = new Map();
            this._outlinedPermanently = null;
            this._outlinedTemporary = null;
            this._svg = null;
            this._anchors = null;
            this._labelTimer = null;
            this._deHighlightTimer = new Map();
            this._label = null;
            this._flyout = null;
            this._zooming = false;
            this._resizeObserver = new ResizeObserver(this.onResize.bind(this));
            this._zoom = d3.zoom().on("zoom", event => { this.transform = event.transform });
        }

        /**
         * Visualize the view box
         * @param {boolean} state
         * @returns {View}
         */
        showWireframe(state) {
            this._showWireframe = state;

            return this;
        }

        /**
         * Provide a callback to render node labels
         *
         * Must be a function that returns a promise that resolves to an HTML node, a string
         * or NULL. The function will be called with the node's data as the first argument.
         *
         * @param {function} label
         * @returns {View}
         */
        label(label) {
            this._label = label;

            return this;
        }

        set box(box) {
            this._box = box;
            this.rescaleZoomConstraints();

            requestAnimationFrame(() => {
                this.svg.attr('viewBox', this._box);

                if (this._showWireframe) {
                    const cells = [];

                    for (let i = this._box[0]; i < this._box[0] + this._box[2]; i += WIREFRAME_CELL_SIZE) {
                        for (let j = this._box[1]; j < this._box[1] + this._box[3]; j += WIREFRAME_CELL_SIZE) {
                            cells.push([ i, j ]);
                        }
                    }

                    this.svg.selectChild(".wireframe")
                        .selectAll("rect")
                        .data(cells, d => d.join(",")) // Use the coordinates as the key
                        .join("rect")
                        .attr("x", d => d[0])
                        .attr("y", d => d[1])
                        .attr("width", WIREFRAME_CELL_SIZE)
                        .attr("height", WIREFRAME_CELL_SIZE)
                        .append("title");
                }

                this.update(); // Fills each title
            });
        }

        get svg() {
            if (this._svg === null) {
                const svg = d3.create("svg").attr('viewBox', this._box);

                if (this._showWireframe) {
                    svg.append("g").classed("wireframe", true);
                }

                const groupLayer = svg.append("g").attr('id', 'group-layer');
                groupLayer.append("g").classed('links', true);
                groupLayer.append("g").classed('nodes', true);
                groupLayer.append("g").classed('labels', true);
                const hierarchyLayer = svg.append("g").attr('id', 'hierarchy-layer');
                hierarchyLayer.append("g").classed('links', true);
                hierarchyLayer.append("g").classed('nodes', true);
                hierarchyLayer.append("g").classed('icons', true);
                hierarchyLayer.append("g").classed('summaries', true);
                const clusterLayer = svg.append("g").attr('id', 'cluster-layer');
                clusterLayer.append("g").classed('links', true);
                clusterLayer.append("g").classed('nodes', true);

                svg.append("g").attr("class", "node-outline");

                this._zoom(svg);

                if (this._label !== null) {
                    svg.node().addEventListener("mouseover", this.onHover.bind(this));
                    svg.node().addEventListener("mouseleave", this.onMapExit.bind(this));
                    this._zoom.on("start", () => (this._zooming = true));
                    this._zoom.on("end", () => (this._zooming = false));
                }

                this._svg = svg;
            }

            return this._svg;
        }

        get flyout() {
            if (this._flyout === null) {
                this._flyout = this.svg.append("g")
                    .attr("class", "node-flyout");
                this._flyout.append("foreignObject")
                    .append("xhtml:div")
                    .classed("node-label", true)
                    .append("xhtml:div")
                    .classed("content", true);
            }

            return this._flyout;
        }

        get anchors() {
            if (this._anchors === null) {
                this._anchors = this.svg
                    .insert("g", "#links")
                    .attr("stroke-width", this.nodeBorderWidth)
                    .attr("stroke-dasharray", "1")
                    .attr("stroke", "lightgrey")
                    .attr("fill", "transparent");
            }

            return this._anchors;
        }

        set anchors(anchors) {
            this.anchors.selectAll("circle")
                .data(anchors, d => d.id)
                .join("circle")
                .attr("cx", d => d.x)
                .attr("cy", d => d.y)
                .attr("r", d => d.r + this.nodeBorderWidth);
        }

        get lod() {
            return this._lod ?? 1;
        }

        set lod(level) {
            this._lod = level;
        }

        /**
         * Get the current transformation
         *
         * @returns {d3.ZoomTransform}
         */
        get transform() {
            return d3.zoomTransform(this.svg.node());
        }

        /**
         * Apply a transformation to all SVG groups
         *
         * @param {d3.ZoomTransform} transform
         */
        set transform(transform) {
            this.lod = this.scaleToLod(transform.k);
            this.update();

            this.updateFlyoutPosition();
            this.updateOutlinePosition();
            this.anchors.attr("transform", transform);
            // Cannot be applied on this.svg, glitches otherwise
            this.svg.selectChildren("[id$='-layer'], [id$='-tree']")
                .attr("transform", transform);
        }

        /**
         * Update the zoom constraints according to the available LOD levels
         * @returns {void}
         */
        updateZoomConstraints() {
            let boundingBox = [ [0, 0], [0, 0] ];
            for (const [ _, data ] of this._lodLevels) {
                data[3].visit((node, x1, y1, x2, y2) => {
                    if (! node.length) {
                        do {
                            boundingBox[0][0] = Math.min(boundingBox[0][0], node.data.x);
                            boundingBox[1][0] = Math.max(boundingBox[1][0], node.data.x);
                            boundingBox[0][1] = Math.min(boundingBox[0][1], node.data.y);
                            boundingBox[1][1] = Math.max(boundingBox[1][1], node.data.y);
                        } while ((node = node.next));
                    }

                    return x1 >= boundingBox[0][0]
                        && y1 >= boundingBox[0][1]
                        && x2 <= boundingBox[1][0]
                        && y2 <= boundingBox[1][1];
                });
            }

            // Cushion the bounding box by 10% on each side but don't shrink it farther than the view box
            boundingBox[0][0] = Math.min(boundingBox[0][0] * 1.1, this._box[0] * 0.9);
            boundingBox[0][1] = Math.min(boundingBox[0][1] * 1.1, this._box[1] * 0.9);
            boundingBox[1][0] = Math.max(boundingBox[1][0] * 1.1, (this._box[0] + this._box[2]) * 0.9);
            boundingBox[1][1] = Math.max(boundingBox[1][1] * 1.1, (this._box[1] + this._box[3]) * 0.9);

            this._zoom.translateExtent(boundingBox);
            this.rescaleZoomConstraints();
        }

        /**
         * Rescale the zoom constraints according to the current view and bounding box
         * @returns {void}
         */
        rescaleZoomConstraints() {
            if (! this.hasPopulation()) {
                return;
            }

            const translateExtent = this._zoom.translateExtent();
            this._zoom.scaleExtent([
                // The scale extent should only allow to zoom out that far that the whole map fits into the view
                Math.min(
                    this._box[2] / (translateExtent[1][0] - translateExtent[0][0]),
                    this._box[3] / (translateExtent[1][1] - translateExtent[0][1])
                ),
                // Zooming in should not be possible farther than the highest LOD plus 1
                Math.max(...this._lodLevels.keys()) + 1
            ]);
        }

        /**
         * Fit the view so that the given nodes are visible
         * @param {Array} nodes If empty, fits the whole map, unless there are highlighted nodes
         * @param {boolean} animate Whether to use a transition
         * @returns {Promise<void>}
         */
        fit(nodes = [], animate = true) {
            if (! nodes.length && this._highlightPermanently.size > 0) {
                // Never fit the whole map if there are highlighted nodes
                return Promise.reject().catch(() => { /* pass */ });
            }

            if (! nodes.length) {
                if (! animate) {
                    this._zoom.scaleTo(this.svg, this._zoom.scaleExtent()[0]);

                    return Promise.resolve();
                }

                return new Promise(resolve => {
                    const transition = this.svg.transition()
                        .duration(2000)
                        .on("end", () => resolve());

                    this._zoom.scaleTo(transition, this._zoom.scaleExtent()[0]);
                });
            }

            let minX = nodes[0].x, // Just so…
                minY = nodes[0].y, // there is…
                maxX = nodes[0].x, // a valid…
                maxY = nodes[0].y; // initial value.
            for (const node of nodes) {
                minX = Math.min(minX, node.x - node.r);
                minY = Math.min(minY, node.y - node.r);
                maxX = Math.max(maxX, node.x + node.r);
                maxY = Math.max(maxY, node.y + node.r);
            }

            const midX = (minX + maxX) / 2;
            const midY = (minY + maxY) / 2;
            const scale = utils.clamp(
                Math.min(
                    this._box[2] / (maxX - minX),
                    this._box[3] / (maxY - minY)
                ),
                ...this._zoom.scaleExtent()
            );

            return this.zoomTo(scale, midX, midY, animate);
        }

        /**
         * Zoom to a specific area
         * @param {Number} scale
         * @param {Number} x
         * @param {Number} y
         * @param {boolean} animate Whether to use a transition
         * @returns {Promise<void>}
         */
        zoomTo(scale, x, y, animate = true) {
            const transform = new d3.ZoomTransform(scale, 0, 0)
                .translate(-x, -y);

            if (! animate) {
                this.svg.call(this._zoom.transform, transform);

                return Promise.resolve();
            }

            return new Promise(resolve => {
                this.svg.transition()
                    .duration(2000)
                    .on("end", () => resolve())
                    .call(this._zoom.transform, transform);
            });
        }

        /**
         * Zoom to a specific node
         * @param {object} node
         * @param {boolean} animate Whether to use a transition
         * @returns {Promise<void>}
         */
        focusNode(node, animate = true) {
            const p = this.zoomTo(Math.max(...this._lodLevels.keys()), node.x, node.y, animate);
            this.highlight([ node.id, ...this.collectRelatedNodes(node.id) ], "magnify");
            this.showOutline(node, true);

            return p;
        }

        /**
         * Remove focus from a node
         * @param {?object} node The node to remove focus from. If null, only the temporary focus is removed.
         */
        blurNode(node = null) {
            this.hideOutline(node, true);
            this.deHighlight("magnify", true);
        }

        /**
         * Convert a transform's scale factor to a valid LOD level
         * @param {Number} scale
         * @returns {Number}
         */
        scaleToLod(scale) {
            const minLevel = Math.min(...this._lodLevels.keys());
            let level = Math.max(minLevel, Math.floor(scale));
            if (! this._lodLevels.has(level)) {
                level = Math.max(1, ...this._lodLevels.keys());
            }

            return level;
        }

        /**
         * Highlight nodes
         * @param {Array<Number>} ids Node IDs to highlight
         * @param {?string} group Group highlighted nodes, allows to differentiate temporary sets. Permanent if null.
         */
        highlight(ids, group = null) {
            if (group === null) {
                this._highlightPermanently.clear();

                for (const id of ids) {
                    this._highlightPermanently.add(id);
                }
            } else {
                if (this._deHighlightTimer.has(group)) {
                    clearTimeout(this._deHighlightTimer.get(group));
                    this._deHighlightTimer.delete(group);
                }

                const set = new Set();
                this._highlightTemporary.set(group, set);

                for (const id of ids) {
                    set.add(id);
                }
            }

            this.update();
        }

        /**
         * Return whether the given node is being highlighted
         * @param {*} id
         * @returns {boolean}
         */
        isHighlighted(id) {
            const highlighted = this._highlightPermanently.has(id)
                || iterator.findKey(this._highlightTemporary, e => e.has(id));

            return highlighted && this.shouldBeActive(id);
        }

        /**
         * Return whether the given node should be active
         * @param {*} id
         * @returns {boolean}
         */
        shouldBeActive(id) {
            const temporary = !! iterator.findKey(this._highlightTemporary, e => e.has(id));
            const others = !! iterator.findKey(this._highlightTemporary, e => e.size > 0);

            return temporary // Temporary? Always.
                || ! others && ( // Not but others are? Never.
                    // Either no permanent ones as well or the node is permanently highlighted.
                    this._highlightPermanently.size === 0 || this._highlightPermanently.has(id)
                );
        }

        /**
         * De-highlight all temporarily highlighted nodes
         *
         * @param {string} group The group the nodes belong to
         * @param {boolean} cancellable Pass true to allow cancelling the de-highlighting
         */
        deHighlight(group, cancellable = false) {
            if (! this._highlightTemporary.has(group)) {
                return;
            }

            const callback = function() {
                const update = this._highlightTemporary.get(group).size > 0;
                this._highlightTemporary.get(group).clear();
                this._deHighlightTimer.delete(group);

                if (update) {
                    this.update();
                }
            }

            if (cancellable) {
                clearTimeout(this._deHighlightTimer.get(group));
                this._deHighlightTimer.set(group, setTimeout(callback.bind(this), 250));
            } else {
                callback.call(this);
            }
        }

        /**
         * Show an outline for the given node
         * @param {object} node
         * @param {boolean} permanent Whether the outline should stay visible until explicitly hidden
         */
        showOutline(node, permanent = false) {
            if (permanent) {
                this._outlinedPermanently = node;
            } else {
                this._outlinedTemporary = node;
            }

            this.svg.select(".node-outline")
                .datum(node)
                .selectAll("path")
                .data(d => [d], d => d.id)
                .join("path")
                .attr("d", d => d3.arc()
                    .innerRadius(d.r + 2)
                    .outerRadius(d.r + 4)
                    .endAngle(2 * Math.PI)
                    .startAngle(0)());

            this.updateOutlinePosition();
        }

        /**
         * Update the outline's position
         */
        updateOutlinePosition() {
            if (this._outlinedPermanently !== null || this._outlinedTemporary !== null) {
                this.svg.select(".node-outline")
                    .attr("transform", d => {
                        // Offset transform by the node's position, path is centered at 0,0
                        const x = this.transform.applyX(d.x);
                        const y = this.transform.applyY(d.y);

                        return `translate(${x},${y}) scale(${this.transform.k})`;
                    });
            }
        }

        /**
         * Hide the outline
         * @param {?object} node The node to hide the outline for
         * @param {boolean} permanent Whether to hide the permanent outline as well
         */
        hideOutline(node = null, permanent = false) {
            if (permanent && this._outlinedPermanently === node) {
                this._outlinedPermanently = null;
            }

            if (permanent || this._outlinedTemporary === node) {
                this._outlinedTemporary = null;
            }

            if (this._outlinedPermanently !== null) {
                this.showOutline(this._outlinedPermanently, true);
            } else {
                this.svg.select(".node-outline").selectAll("path").remove();
            }
        }

        /**
         * Show a flyout with content at the given node
         *
         * @param {SVGCircleElement} node
         * @param {HTMLElement|string} content
         */
        showFlyout(node, content) {
            const nodeRect = node.getBoundingClientRect();
            const containerRect = this.svg.node().getBoundingClientRect();
            const nodeCenterX = nodeRect.left + (nodeRect.width / 2);
            const nodeCenterY = nodeRect.top + (nodeRect.height / 2);
            const cx = node.cx.baseVal.value;
            const cy = node.cy.baseVal.value;
            const data = d3.select(node).datum();

            this.showOutline(data);

            if (content instanceof HTMLElement) {
                this.flyout.selectAll('.content').append(() => content);
            } else {
                this.flyout.selectAll('.content').append("p").text(content);
            }

            const nodeLabel = this.flyout.node().querySelector(".node-label");
            const nodeLabelRect = nodeLabel.getBoundingClientRect();

            this.flyout.select("foreignObject")
                .attr("width", nodeLabelRect.width)
                .attr("height", nodeLabelRect.height);

            let flyoutV = "bottom";
            let cyModifier = 0;
            if (nodeCenterY + nodeLabelRect.height > containerRect.bottom) {
                cyModifier = -nodeLabelRect.height;
                flyoutV = "top";
            }

            nodeLabel.classList.add(flyoutV);

            let flyoutH = "center";
            let cxModifier = 0;
            if (nodeCenterX + (nodeLabelRect.width / 2) > containerRect.right) {
                flyoutH = "left";
                cxModifier = -nodeLabelRect.width;
            } else if (nodeCenterX - (nodeLabelRect.width / 2) < containerRect.left) {
                flyoutH = "right";
            } else {
                cxModifier = -(nodeLabelRect.width / 2);
            }

            nodeLabel.classList.add(flyoutH);

            this.flyout.datum({
                node_id: data.id,
                xm: cxModifier,
                ym: cyModifier,
                x: cx,
                y: cy
            });

            this.updateFlyoutPosition();
        }

        /**
         * Update the flyout position, to keep it in sync with the node
         */
        updateFlyoutPosition() {
            if (this._flyout !== null) {
                // transform is only applied to coordinates because the HTML must not be scaled
                this.flyout.select("foreignObject")
                    .attr("x", d => this.transform.applyX(d.x) + d.xm)
                    .attr("y", d => this.transform.applyY(d.y) + d.ym);
            }
        }

        /**
         * Hide the flyout
         */
        hideFlyout() {
            if (this._flyout !== null) {
                this._flyout.remove();
                this._flyout = null;
            }
        }

        /**
         * Request a label for the given node
         * @param {object} node
         * @returns {Promise<HTMLElement|string|null>}
         */
        async requestLabel(node) {
            if (node.type === "group") {
                if (node.whole) {
                    return null;
                }

                const list = document.createElement("ul");
                for (const group of node.id.split("|")) {
                    list.appendChild(document.createElement("li")).textContent = group || "n/a";
                }

                return list;
            }

            return this._label(node);
        }

        async update() {
            if (! this._lodLevels.has(this.lod)) {
                return;
            }

            const activeLayer = this._lodLevels.get(this.lod)[0];
            this.svg.attr("class", activeLayer + "-layer");

            let toRender = [];
            switch (activeLayer) {
                case "cluster":
                    toRender.push("cluster");

                    // Fallthrough
                case "hierarchy":
                    toRender.push("hierarchy");

                    break; // TODO: Fallthrough, once groups should be always visible
                case "group":
                    toRender.push("group");
            }

            const promises = [];
            // TODO: I tried my best to avoid any assumptions about the order of layers or their presence
            //  but now I give up as time (and motivation) runs out…
            for (const layer of [ "group", "hierarchy", "cluster" ]) {
                const lod = iterator.findKey(this._lodLevels, e => e[0] === layer);
                if (typeof lod !== "undefined" && toRender.includes(layer)) {
                    promises.push(this.getVisibleNodes(lod).then(([nodes, links, labels]) => {
                        return new Promise(resolve => {
                            requestAnimationFrame(() => {
                                this.renderLayer(this.svg.selectChild(`#${layer}-layer`), nodes, links, labels);
                                resolve();
                            });
                        });
                    }));
                } else {
                    promises.push(new Promise(resolve => {
                        requestAnimationFrame(() => {
                            this.renderLayer(this.svg.selectChild(`#${layer}-layer`), [], [], []);
                            resolve();
                        });
                    }));
                }
            }

            if (this._showWireframe) {
                promises.push(new Promise(resolve => {
                    requestAnimationFrame(() => {
                        this.svg.selectAll(".wireframe rect")
                            .select("title")
                            .text(d => {
                                const [ rx, ry ] = this.transform.apply(d);

                                return `${d[0]},${d[1]} (${rx},${ry})`;
                            });

                        resolve();
                    });
                }));
            }

            return Promise.all(promises);
        }

        getVisibleNodes(lod) {
            return new Promise((resolve, reject) => {
                setTimeout(() => {
                    const [ _, nodes, links, tree ] = this._lodLevels.get(lod);
                    const visibleNodes = utils.findVisibleNodes(tree, this.transform, 0.02, ...this._box);

                    const newLinks = [];
                    for (const link of links.values()) {
                        const hasSource = visibleNodes.has(link.source_id);
                        const hasTarget = visibleNodes.has(link.target_id);
                        if (hasSource || hasTarget) {
                            newLinks.push(link);
                            if (! hasTarget) {
                                visibleNodes.add(link.target_id);
                            } else if (! hasSource) {
                                visibleNodes.add(link.source_id);
                            }
                        }
                    }

                    const newNodes = [];
                    for (const nodeId of visibleNodes) {
                        newNodes.push(nodes.get(nodeId));
                    }

                    const nodesWithLabel = newNodes.filter(d => d.type === "group" || d.type === "blob");

                    resolve([ newNodes, newLinks, nodesWithLabel ]);
                }, 0);
            });
        }

        renderLayer(layer, nodes, links, labels) {
            layer.selectChild('.nodes').selectChildren()
                .data(nodes, d => d.id)
                .join(
                    enter => enter.append("circle")
                        .attr("cx", d => d.x)
                        .attr("cy", d => d.y)
                        .attr("r", d => d.r + this.nodeBorderWidth),
                    update => {
                        if (this._zooming) {
                            return update; // Positions do not change while zooming/panning
                        }

                        return update
                            .attr("cx", d => d.x)
                            .attr("cy", d => d.y)
                            .attr("r", d => d.r + this.nodeBorderWidth);
                    },
                    exit => {
                        if (this._flyout !== null) {
                            for (const node of exit) {
                                const data = d3.select(node).datum();
                                if (this._flyout.datum().node_id === data.id) {
                                    this.hideOutline(data, true);
                                    this.hideFlyout();

                                    break;
                                }
                            }
                        }

                        exit.remove();
                    }
                )
                .attr("class", d => {
                    const classes = [ d.type ];

                    if (this.shouldBeActive(d.id)) {
                        if ("state" in d) {
                            classes.push(`state-${d.state}`);
                        } else if ("link" in d && "failed" in d.link) {
                            classes.push(`state-${d.link.failed}`);
                        }
                    }

                    return classes.join(" ");
                })
                .classed("highlighted", d => this.isHighlighted(d.id))
                .classed("node", d => d.type !== "blob" && d.type !== "group")
                .classed("cluster", d => d.type === "group" || d.type === "blob");

            layer.selectChild('.links').selectChildren()
                .data(links, d => d.id)
                .join(
                    enter => enter.append("polygon")
                        .attr("points", d => d.points.map(p => `${p.x},${p.y}`).join(" "))
                        .attr("x2", d => d.x2)
                        .attr("y2", d => d.y2)
                        .attr("x1", d => d.x1)
                        .attr("y1", d => d.y1),
                    update => {
                        if (this._zooming) {
                            return update;
                        }

                        return update
                            .attr("points", d => d.points.map(p => `${p.x},${p.y}`).join(" "))
                            .attr("x2", d => d.x2)
                            .attr("y2", d => d.y2)
                            .attr("x1", d => d.x1)
                            .attr("y1", d => d.y1);
                    },
                )
                .attr("class", d => {
                    if (this.shouldBeActive(d.source_id) && this.shouldBeActive(d.target_id)) {
                        return "failed" in d ? `state-${d.failed}` : null;
                    }
                })
                .classed(
                    "highlighted",
                    d => this.isHighlighted(d.source_id) && this.isHighlighted(d.target_id)
                );

            // Group labels
            layer.selectChild('.labels').selectChildren("foreignObject")
                .data(() => {
                    const groups = [];
                    const scale = this.transform.k;
                    for (const node of labels) {
                        // The following is based on an average of 8px width and 16px height per character
                        const candidates = node.id.split("|");
                        const radius = scale * node.r;
                        let names = [];
                        let whole = true;
                        if (radius > 16) {
                            // The names are centered at the node's position, but the diameter varies with the order
                            const diameters = [
                                Math.sqrt(radius ** 2 - 16 ** 2) * 2, // The first name
                                scale * node.r * 2, // The second
                                Math.sqrt(radius ** 2 - 16 ** 2) * 2 // The third
                            ];

                            for (let name of candidates) {
                                if (! diameters.length) {
                                    whole = false;

                                    break;
                                }

                                if (! name) {
                                    name = "n/a";
                                }

                                const diameter = diameters.pop();
                                if (name.length * 8 < diameter) {
                                    names.push(name);
                                } else {
                                    whole = false;
                                    diameters.push(diameter);
                                }
                            }
                        }

                        if (! whole && scale * node.r < 32) {
                            // There is not enough space for +n and the names
                            names = [];
                        }

                        if (names.length) {
                            node.whole = whole;
                            if (! whole) {
                                names.push(`+${candidates.length - names.length}`);
                            }

                            groups.push({ id: node.id, x: node.x, y: node.y, names });
                        } else if (scale * node.r >= 16) {
                            groups.push({ id: node.id, x: node.x, y: node.y, names: [ `+${candidates.length}` ] });
                        }
                    }

                    return groups;
                }, d => d.id)
                .join(
                    enter => {
                        enter = enter.append("foreignObject")
                            // TODO: Invisible in Safari, should have a proper width and height
                            .style("overflow", "visible")
                            .attr("x", d => d.x)
                            .attr("y", d => d.y)
                            .attr("width", 1)
                            .attr("height", 1)
                            .classed("group-label", true);
                        enter.append("xhtml:ul");

                        return enter;
                    }
                )
                .selectChild("ul")
                .style("transform", _ => `translate(-50%, -50%) scale(${1 / this.transform.k})`)
                .selectChildren("li")
                .data(d => d.names, d => d)
                .join("li")
                .text(d => d);

            // Render an unreachable icon for blobs with failed links
            layer.selectChild('.icons').selectChildren("text")
                .data(labels.filter(d => d.type === "blob" && d?.link?.failed === 1), d => d.id)
                .join(
                    enter => enter.append("text")
                        .attr("x", d => d.x)
                        .attr("y", d => d.y)
                        .attr("text-anchor", "middle")
                        .attr("dominant-baseline", "middle")
                        .classed("blob-label fa", true)
                        .text("\uf0e8")
                );

            if (this._blobSummaries) {
                layer.selectChild('.summaries').selectChildren("foreignObject")
                    .data(labels.filter(d => d.type === "blob" && d?.link?.failed !== 1), d => d.id)
                    .join(
                        enter => {
                            enter = enter.append("foreignObject")
                                // TODO: Invisible in Safari, should have a proper width and height
                                .style("overflow", "visible")
                                .attr("x", d => d.x)
                                .attr("y", d => d.y)
                                .attr("width", 1)
                                .attr("height", 1);
                            enter.append("xhtml:ul")
                                .attr("class", "cluster-summary");

                            return enter;
                        }
                    )
                    .selectChild("ul")
                    .selectChildren("li")
                    .data(d => {
                        const stateSummary = new Map();
                        for (const member of d.members) {
                            if (typeof member === "object" && "state" in member) {
                                const key = `${member.type}-${member.state}`;
                                let currentSum = stateSummary.get(key);
                                if (typeof currentSum === "undefined") {
                                    currentSum = {count: 0, type: member.type, state: member.state};
                                    stateSummary.set(key, currentSum);
                                }

                                currentSum.count++;
                            }
                        }

                        if (stateSummary.size === 0) {
                            return [];
                        }

                        return [...stateSummary.values()];
                    }, d => `${d.type}-${d.state}-${d.count}`)
                    .join(
                        enter => enter.append("li")
                            .text(d => d.count)
                    )
                    .attr("class", d => {
                        const classes = [ d.type ];
                        if (this.shouldBeActive(d.id)) {
                            classes.push(`state-${d.state}`);
                        }

                        return classes.join(" ");
                    });
            } else if (this._blobDonuts) {
                layer.selectChild('.summaries').selectChildren("g")
                    .data(labels.filter(d => d.type === "blob" && d?.link?.failed !== 1), d => d.id)
                    .join(
                        enter => enter.append("g")
                            .classed("cluster-donut", true)
                            .attr("transform", d => `translate(${d.x},${d.y})`)
                    )
                    .selectAll("path")
                    .data(d => {
                        const stateSummary = new Map();
                        for (const member of d.members) {
                            if (typeof member === "object" && "state" in member) {
                                const key = `${member.type}-${member.state}`;
                                let currentSum = stateSummary.get(key);
                                if (typeof currentSum === "undefined") {
                                    currentSum = { count: 1, type: member.type, state: member.state };
                                    stateSummary.set(key, currentSum);
                                }

                                currentSum.count++;
                            }
                        }

                        if (stateSummary.size === 0) {
                            return [];
                        }

                        // Half offsets to make the arcs look like strokes
                        const arc = d3.arc().innerRadius(d.r + .5).outerRadius(d.r + 1.5);
                        const pie = d3.pie().value(d => d.count);

                        const arcs = [];
                        for (const angle of pie(stateSummary.values())) {
                            arcs.push({
                                blobId: d.id,
                                arc: arc(angle),
                                type: angle.data.type,
                                state: angle.data.state,
                                count: angle.data.count
                            });
                        }

                        return arcs;
                    }, d => `${d.type}-${d.state}-${d.count}`)
                    .join(
                        enter => enter.append("path")
                            .attr("d", d => d.arc)
                    )
                    .attr("class", d => {
                        const classes = [ d.type ];
                        if (this.shouldBeActive(d.blobId)) {
                            classes.push(`state-${d.state}`);
                        }

                        return classes.join(" ");
                    });
            }
        }

        renderTree(tree, layer) {
            const nodes = [];
            tree.visit((node, x1, y1, x2, y2) => {
                if (node.length) {
                    nodes.push([ x1, y1, x2 - x1, y2 - y1 ]);
                }
            });

            let group = this.svg.selectChild(`#${layer}-tree`);
            if (group.empty()) {
                group = this.svg.insert("g", ".wireframe")
                    .attr("id", `${layer}-tree`)
                    .classed("quadtree", true);
            }

            group.selectAll("rect")
                .data(nodes, d => d.join(","))
                .join("rect")
                .attr("x", d => d[0])
                .attr("y", d => d[1])
                .attr("width", d => d[2])
                .attr("height", d => d[3]);
        }

        async populate(layer, nodes, links) {
            const nodesMap = new Map(nodes.map(d => [ d.id, d ]));
            const linkMap = new Map(links.map(d => [ d.id, d ]));
            const tree = d3.quadtree(nodes, d => d.x, d => d.y);

            let level = iterator.findKey(this._lodLevels, e => e[0] === layer);
            const update = typeof level !== "undefined";
            if (! update) {
                // TODO: Again, so much about not making assumptions about the order of layers…
                if (layer === "group" && this._lodLevels.size > 0) {
                    level = 1;
                    this._lodLevels.clear();
                } else {
                    level = Math.max(...this._lodLevels.keys());
                    if (level === -Infinity) {
                        level = 1;
                    } else {
                        level++;
                    }
                }
            }

            if (update && layer === "group") {
                // Do a full reset
                this._lodLevels.clear();
            } else if (update && layer === "hierarchy") {
                // Empty the cluster layer, if it exists
                const clusterLod = iterator.findKey(this._lodLevels, e => e[0] === "cluster");
                if (clusterLod) {
                    this._lodLevels.delete(clusterLod);
                }
            } else if (layer === "cluster") {
                // Resolve blob member ids to data, once
                const hierarchyLevel = iterator.find(this._lodLevels.values(), e => e[0] === "hierarchy");
                const blobs = iterator.filter(hierarchyLevel[1].values(), d => d.type === "blob");
                for (const blob of blobs) {
                    blob.members = blob.members.map(id => typeof id !== "object" ? nodesMap.get(id) : id);
                }
            }

            this._lodLevels.set(level, [ layer, nodesMap, linkMap, tree ]);
            this.updateZoomConstraints();

            if (this._showWireframe) {
                this.renderTree(tree, layer);
            }

            const potentialLod = this.scaleToLod(this.transform.k);
            if (level === potentialLod) {
                this.lod = level;

                return this.update();
            } else if (update || (layer === "cluster" && this.lod === level - 1)) {
                return this.update();
            }
        }

        /**
         * Count the number of layers
         * @returns {Number}
         */
        countLayers() {
            return this._lodLevels.size;
        }

        hasPopulation() {
            return this._lodLevels.size > 0;
        }

        clearLayer(layer) {
            const level = iterator.findKey(this._lodLevels, e => e[0] === layer);
            if (typeof level !== "undefined") {
                this._lodLevels.delete(level);

                const currentLod = this.lod;
                if (this._lodLevels.has(level + 1)) {
                    // Re-index the following layers
                    const lodLevels = new Map();
                    for (const [ lod, data ] of this._lodLevels) {
                        if (lod > level) {
                            lodLevels.set(lod - 1, data);

                            if (lod === currentLod) {
                                this.lod = lod - 1;
                                this.update();
                            }
                        } else {
                            lodLevels.set(lod, data);
                        }
                    }

                    this._lodLevels = lodLevels;
                } else if (currentLod === level) {
                    this.lod = level - 1;
                    this.update();
                }

                this.updateZoomConstraints();
            }
        }

        bindTo(container) {
            const svgNode = this._svg?.node();
            if (typeof svgNode !== 'undefined') {
                this._resizeObserver.unobserve(svgNode);
            }

            this._resizeObserver.observe(this.svg.node());
            container.appendChild(this.svg.node());

            return this;
        }

        onResize(entries) {
            const entry = entries[0]; // There's only one map-container
            if (! entry.target.isConnected) {
                return;
            }

            let width, height;
            if (entry.contentBoxSize) {
                if (entry.contentBoxSize[0]) {
                    // The standard makes contentBoxSize an array...
                    width = entry.contentBoxSize[0].inlineSize;
                    height = entry.contentBoxSize[0].blockSize;
                } else {
                    // ...but old versions of Firefox treat it as a single item
                    width = entry.contentBoxSize.inlineSize;
                    height = entry.contentBoxSize.blockSize;
                }
            } else {
                // Some ancient browsers don't support contentBoxSize, kept for compatibility
                width = entry.contentRect.width;
                height = entry.contentRect.height;
            }

            this.box = [-width / 2, -height / 2, width, height];
        }

        onHover(event) {
            if (this._zooming) {
                return;
            }

            if (this._labelTimer !== null) {
                // mouseover is only triggered on entry, so if a timer is running, it's for another node
                clearTimeout(this._labelTimer);
                this._labelTimer = null;
            }

            event.stopPropagation(); // We know what to do afterward, no need to decide again

            if (this._flyout !== null && this.flyout.node().contains(event.target)) {
                // The user moved the mouse from the node to the flyout
                return;
            } else if (event.target.tagName.toLowerCase() !== "circle") {
                // Not a node nor something inside the flyout?
                this.hideFlyout();
                this.hideOutline();
                this.deHighlight("hover", true);

                return;
            }

            const data = d3.select(event.target).datum();
            if (this._flyout !== null) {
                // It might be the node we're already showing the flyout for
                if (this.flyout.datum().node_id === data.id) {
                    return;
                }

                // Or some other
                this.hideFlyout();
            }

            // Cancel any potential de-highlighting to avoid flickering
            if (this._deHighlightTimer.has("hover")) {
                clearTimeout(this._deHighlightTimer.get("hover"));
                this._deHighlightTimer.delete("hover");
            }

            const node = event.target;
            const timer = setTimeout(() => {
                this.deHighlight("hover");

                if (this._zooming) {
                    this.hideOutline(); // Hide outline only if there is no other node to be outlined

                    return;
                }

                this.requestLabel(data).then(label => {
                    if (this._labelTimer !== timer || this._zooming) {
                        // The timer was reset, so the user moved the mouse away
                        // or the user zoomed in/out
                    } else if (label !== null) {
                        this.showFlyout(node, label);

                        return;
                    }

                    this.hideOutline();
                });

                this.highlight([ data.id, ...this.collectRelatedNodes(data.id) ], "hover");
            }, 250);
            this._labelTimer = timer;
        }

        onMapExit(event) {
            clearTimeout(this._labelTimer);
            this.hideFlyout();
            this.hideOutline();
            this.deHighlight("hover");
        }

        /**
         * Collect all related nodes of the given node
         * @param {*} id
         * @param {boolean} parents
         * @param {Map} nodes
         * @param {Map} edges
         * @returns {Generator<*>}
         */
        *collectRelatedNodes(id, parents = true, nodes = null, edges = null) {
            let data;
            let optionalChildren = parents;
            if (nodes === null && edges === null) {
                for (const [ layer, n, e, _ ] of this._lodLevels.values()) {
                    if (n.has(id)) {
                        data = n.get(id);
                        if (layer === "cluster") {
                            [ nodes, edges ] = iterator.find(this._lodLevels.values(), e => e[0] === "hierarchy")
                                .slice(1, 3);
                        } else {
                            nodes = n;
                            edges = e;
                        }

                        break;
                    }
                }
            } else {
                data = nodes.get(id);
                optionalChildren = false;
            }

            if (typeof data === "undefined") {
                throw new Error("Node not found");
            } else if (data.type === "blob") {
                // I assume it's fine to yield the blob members.
                // Since blobs have no children themselves, they can only be
                // reached by the optionalChildren path anyway. (or if hovered)
                for (const member of data.members) {
                    if (isNaN(member)) {
                        yield member.id;
                    } else {
                        yield member;
                    }
                }
            }

            if (parents) {
                for (const edgeId of data.parentEdges ?? []) {
                    const edge = edges.get(edgeId);
                    if (edge.source_id !== id) {
                        // The edge is that of a blob
                        yield edge.source_id;
                    }

                    yield edge.target_id;
                    yield* this.collectRelatedNodes(edge.target_id, parents, nodes, edges);
                }
            }

            if (optionalChildren || ! parents) {
                for (const edgeId of data.childEdges ?? []) {
                    const edge = edges.get(edgeId);

                    yield edge.source_id;
                    yield* this.collectRelatedNodes(edge.source_id, false, nodes, edges);
                }
            }
        }

        reset(full = true) {
            if (full) {
                const svgNode = this._svg?.node();
                if (typeof svgNode !== 'undefined') {
                    this._resizeObserver.unobserve(svgNode);
                    svgNode.remove();
                }

                this._svg = null;
                this._label = null;
            }

            this._lod = null;
            this._lodLevels.clear();
            this._highlightPermanently.clear();
            this._highlightTemporary.clear();
            this._outlinedPermanently = null;
            this._outlinedTemporary = null;
            this._anchors = null;
            this._labelTimer = null;
            this._deHighlightTimer.clear();
            this._flyout = null;
            this._zooming = false;
        }
    }

    return View;
});
