/* Icinga Dependencies Web | (c) 2024 Icinga GmbH | AGPLv3 */

"use strict";

const NODE_RADIUS_MIN = 2;
const NODE_RADIUS_MAX = 10;
// Why double border? Because SVG stroke is centered on the path and browsers don't render half pixels
const NODE_BORDER_WIDTH = 2;

function identifyClusters(nodes, edges) {
    let groupedNodes = d3.group(nodes, ([_, { id }]) => {
        const parents = new Set(icingaIteratorPolyfill.map(
            edges.get(id).get("parents"), ([ parentId, { stateId } ]) => parentId + '|' + stateId));
        const children = new Set(icingaIteratorPolyfill.map(
            edges.get(id).get("children"), ([ childId, { stateId } ]) => childId + '|' + stateId));

        let group = [...parents.values()].sort().toString();
        if (children.size > 0) {
            if (group) {
                group += '|';
            }

            group += [...children.values()].sort().toString();
        }

        return group;
    });
    // TODO: only commented as I'm not sure it wont be needed later on
    /*if (groupedNodes.size / relatedNodes.size > 0.1) {
        const reGroupedNodes = d3.group(groupedNodes.keys(), group => {
            const members = groupedNodes.get(group);
            if (members.length === 1) {
                const [nodeId, dependencyData] = members[0];
                const edge = edges.get(nodeId);
                if (edge.get("children").size > 0 || edge.get("parents").size !== 2) {
                    return group;
                } else {
                    const [_, { label: oppositeDependencyLabel }] = d3.reduce(edge.get("parents"), (acc, data) => {
                        return data[0] !== id ? data : acc;
                    }, [null, { label: null }]);
                    return dependencyData.label + '|' + oppositeDependencyLabel;
                }
            }

            return group;
        });

        groupedNodes = function* (base) {
            for (const [_, groups] of reGroupedNodes) {
                const members = [];
                for (const group of groups) {
                    members.push(...base.get(group));
                }

                yield [_, members];
            }
        }(groupedNodes);
    }*/

    for (const [ _, members ] of groupedNodes) {
        if (members.length > 1) {
            let blobId = null;
            do {
                blobId = icingaDependencyMapUtils.generateId(8);
            } while (nodes.has(blobId));

            // All members have the same radius, so we can just take the first one
            const diameter = ((members[0][1].r + NODE_BORDER_WIDTH) * 2) * (Math.sqrt(members.length) + 1);

            nodes.set(blobId, {
                id: blobId,
                type: "blob",
                r: (diameter * 1.2) / 2
            });

            const blobParents = new Map();
            const blobChildren = new Map();
            const groups = new Set();
            const memberIds = [];
            for (const [ nodeId, __ ] of members) {
                nodes.get(nodeId).group.forEach(c => groups.add(c));

                const edge = edges.get(nodeId);
                edge.get("parents").forEach((d, id) => {
                    if (! edges.has(id)) {
                        const parentBlobData = nodes.get(nodes.get(id).blobId);
                        if (! blobParents.has(parentBlobData.id)) {
                            blobParents.set(parentBlobData.id, d);
                        }
                    } else {
                        const parentChildren = edges.get(id).get("children");
                        parentChildren.delete(nodeId);
                        if (! parentChildren.has(blobId)) {
                            parentChildren.set(blobId, d);
                        }

                        if (! blobParents.has(id)) {
                            blobParents.set(id, d);
                        }
                    }
                });
                edge.get("children").forEach((d, id) => {
                    if (! edges.has(id)) {
                        const childBlobData = nodes.get(nodes.get(id).blobId);
                        if (! blobChildren.has(childBlobData.id)) {
                            blobChildren.set(childBlobData.id, d);
                        }
                    } else {
                        const childParents = edges.get(id).get("parents");
                        childParents.delete(nodeId);
                        if (! childParents.has(blobId)) {
                            childParents.set(blobId, d);
                        }

                        if (! blobChildren.has(id)) {
                            blobChildren.set(id, d);
                        }
                    }
                });

                memberIds.push(nodeId);
                edges.delete(nodeId);
                nodes.get(nodeId).blobId = blobId;
            }

            Object.assign(nodes.get(blobId), {
                group: [ ...groups ],
                members: memberIds
            });
            edges.set(blobId, new Map([
                ["parents", blobParents],
                ["children", blobChildren]
            ]));
        }
    }
}

function scaleHierarchy(nodeId, nodes, edges, depth = 1, scale = null) {
    if (scale === null) {
        scale = d3.scaleSqrt([depth + 1, depth], [NODE_RADIUS_MIN, NODE_RADIUS_MAX]);
    } else {
        scale = scale.copy().domain([depth, scale.domain()[1]]);
    }

    let finalScale = scale;
    for (const [ childId, _ ] of edges.get(nodeId).get("children")) {
        const newScale = scaleHierarchy(childId, nodes, edges, depth + 1, scale);
        if (newScale.domain()[0] > finalScale.domain()[0]) {
            finalScale = newScale;
        }
    }

    const data = nodes.get(nodeId);
    if (! ("r" in data)) {
        data.r = finalScale(depth);
    }

    return finalScale;
}

async function parseStructure(stream) {
    const nodes = new Map();
    const edges = new Map();
    const suspiciousEdges = new Set();
    for await (const line of icingaDependencyMapUtils.readLines(stream)) {
        const data = JSON.parse(line);
        if (data[0] === 1) { // node
            let [ _, id, group = "" ] = data;
            if (group.includes("|")) {
                group = group.split("|");
            } else {
                group = [group];
            }

            nodes.set(id, { id, group });
            edges.set(id, new Map([
                ["parents", new Map()],
                ["children", new Map()]
            ]));
        } else if (data[0] === 2) { // edge
            const [ _, id, stateId, source, target ] = data;

            // Edges without a known source and target are ignored. There is no need to show
            // them in the graph. Though, if either end is known, the edge is tracked, but
            // inspected at the end so that only suspicious parents are kept.
            if (! edges.has(source) && ! edges.has(target)) {
                continue;
            } else if (! edges.has(source)) {
                suspiciousEdges.add(source);
                edges.set(source, new Map([
                    ["parents", new Map()],
                    ["children", new Map()]
                ]));
            } else if (! edges.has(target)) {
                suspiciousEdges.add(target);
                edges.set(target, new Map([
                    ["parents", new Map()],
                    ["children", new Map()]
                ]));
            }

            edges.get(source).get("parents").set(target, { id, stateId });
            edges.get(target).get("children").set(source, { id, stateId });
        } else if (data[0] === 3) { // error
            const [ _, message, cause ] = data;
            throw new Error(message, { cause });
        }
    }

    // Only keep suspicious edges that have children themselves
    for (const id of suspiciousEdges) {
        if (edges.get(id).get("children").size > 0) {
            nodes.set(id, { id, group: [ "" ] });
        } else {
            for (const [ parentId, _ ] of edges.get(id).get("parents")) {
                edges.get(parentId).get("children").delete(id);
            }

            edges.delete(id);
        }
    }

    for (const [ nodeId, edge ] of edges) {
        if (edge.get("parents").size === 0) {
            scaleHierarchy(nodeId, nodes, edges);
        }
    }

    identifyClusters(nodes, edges);

    return [ nodes, edges ];
}

function breakupNodes(nodes, edges) {
    const hierarchyNodes = [];
    const clusterMap = new Map();
    for (const [ id, data ] of nodes) {
        if ("blobId" in data) {
            if (! clusterMap.has(data.blobId)) {
                clusterMap.set(data.blobId, new Set());
            }

            clusterMap.get(data.blobId).add(id);
            delete data.blobId; // We don't need this anymore
        } else {
            hierarchyNodes.push(data);
        }
    }

    const hierarchyEdges = [];
    for (const [ nodeId, edge ] of edges) {
        for (const [ childId, d ] of edge.get("children")) {
            hierarchyEdges.push({
                id: d.id,
                stateId: d.stateId,
                source: childId,
                source_id: childId,
                target: nodeId,
                target_id: nodeId
            });
        }
    }

    return [ hierarchyNodes, hierarchyEdges, clusterMap ];
}

function* prepareLayers(nodes, edges) {
    yield { networkIdle: true };

    const [ hierarchyNodes, hierarchyEdges, clusterMap ] = breakupNodes(nodes, edges);

    const groupForce = icingaDependencyMapGroupForce(hierarchyEdges).radius(d => d.r + NODE_BORDER_WIDTH);
    const simulation = d3.forceSimulation(hierarchyNodes)
        .force("group", groupForce)
        .stop();

    let overallTicks = 0;
    const totalTicks = Math.ceil(Math.log(simulation.alphaMin()) / Math.log(1 - simulation.alphaDecay()));

    yield { totalTicks: totalTicks + clusterMap.size };

    for (let i = 0; i < totalTicks; i++) {
        simulation.tick();

        yield { tick: ++overallTicks };
    }

    const groups = groupForce.groups();
    for (const group of groups) {
        // Remove properties defined by the force simulation
        delete group.index;
        delete group.vx;
        delete group.vy;

        // Round the coordinates to reduce the size of the JSON
        group.x = icingaDependencyMapUtils.roundNumber(group.x);
        group.y = icingaDependencyMapUtils.roundNumber(group.y);
        group.r = icingaDependencyMapUtils.roundNumber(group.r);
    }

    const groupLayer = {
        layer: "group",
        nodes: groups,
        links: []
    };

    yield groupLayer;

    const scale = d3.scaleLinear([NODE_RADIUS_MIN, NODE_RADIUS_MAX], [1, 3]).clamp(true);
    for (const link of hierarchyEdges) {
        link.points = icingaDependencyMapUtils.perpendicularPoints(
            { ...link.target, r: link.target.r + NODE_BORDER_WIDTH },
            { ...link.source, r: link.source.r + NODE_BORDER_WIDTH },
            scale(link.target.r),
            0.5
        );

        link.target.childEdges = link.target?.childEdges || [];
        link.target.childEdges.push(link.id);
        link.source.parentEdges = link.source?.parentEdges || [];
        link.source.parentEdges.push(link.id);

        if (link.source.type === "blob" || link.target.type === "blob") {
            link.blob = true;
        }

        delete link.index;
        delete link.vx;
        delete link.vy;
        delete link.source;
        delete link.target;
    }

    for (const node of hierarchyNodes) {
        delete node.index;
        delete node.vx;
        delete node.vy;
        delete node.xOffset;
        delete node.yOffset;
        delete node.group;

        node.x = icingaDependencyMapUtils.roundNumber(node.x);
        node.y = icingaDependencyMapUtils.roundNumber(node.y);
        node.r = icingaDependencyMapUtils.roundNumber(node.r);
    }

    const hierarchyLayer = {
        layer: "hierarchy",
        nodes: hierarchyNodes,
        links: hierarchyEdges
    };

    yield hierarchyLayer;

    const clusterNodes = [];
    for (const [ blobId, memberIds ] of clusterMap) {
        const members = [ ...icingaIteratorPolyfill.filter(nodes.values(), d => memberIds.has(d.id)) ];
        const blobData = nodes.get(blobId);
        const simulation = d3.forceSimulation(members)
            .alphaDecay(1 - (0.001 ** (1 / 100))) // 100 ticks
            .force("collide", d3.forceCollide().strength(2).radius(d => d.r + NODE_BORDER_WIDTH).iterations(2))
            .force("pull", d3.forceManyBody().strength(10).distanceMin((members[0].r + NODE_BORDER_WIDTH) * 2))
            .stop();

        const totalTicks = Math.ceil(Math.log(simulation.alphaMin()) / Math.log(1 - simulation.alphaDecay()));
        for (let i = 0; i < totalTicks; i++) {
            simulation.tick();
        }

        const parents = edges.get(blobId).get("parents");
        for (const member of members) {
            member.parentEdges = member.parentEdges || [];
            for (const [ _, edge ] of parents) {
                member.parentEdges.push(edge.id);
            }

            delete member.index;
            delete member.vx;
            delete member.vy;
            delete member.group;

            // Cluster nodes are closer together, so it's not possible to remove all decimal places
            member.x = icingaDependencyMapUtils.roundNumber(member.x, 2);
            member.y = icingaDependencyMapUtils.roundNumber(member.y, 2);
            member.r = icingaDependencyMapUtils.roundNumber(member.r, 2);

            member.x += blobData.x;
            member.y += blobData.y;
            clusterNodes.push(member);
        }

        yield { tick: ++overallTicks };
    }

    const clusterLayer = {
        layer: "cluster",
        nodes: clusterNodes,
        links: []
    };

    yield clusterLayer;

    return [
        groupLayer,
        hierarchyLayer,
        clusterLayer
    ];
}

async function main(csrfToken, options) {
    if (typeof d3 === 'undefined') {
        let requestD3 = false;
        try {
            importScripts(options.baseUrl + "/lib/icinga/icinga-php-thirdparty/js/mbostock/d3.min.js");
        } catch (error) {
            requestD3 = true;
        }

        importScripts(options.baseUrl + "/dependencies/map/worker-dependencies" + (requestD3 ? "?d3" : ""));
    }

    let cacheUrl = options.baseUrl + "/dependencies/map/cache";
    let dataUrl = options.baseUrl + "/dependencies/map/structure";
    if ("group" in options && options.group) {
        dataUrl += "?groupBy=" + options.group;
        cacheUrl += "?groupBy=" + options.group + "&csrfToken=" + csrfToken;
        if ("invalidate" in options && options.invalidate) {
            dataUrl += "&invalidate";
        }
    } else if ("invalidate" in options && options.invalidate) {
        dataUrl += "?invalidate";
        cacheUrl += "?csrfToken=" + csrfToken;
    } else {
        cacheUrl += "?csrfToken=" + csrfToken;
    }

    const response = await fetch(dataUrl);
    if (response.status !== 200) {
        throw new Error(
            'Failed to fetch dependency map structure: ' + response.statusText,
            { cause: new Error(await response.text()) }
        );
    }

    let preparer;
    if (response.headers.get("Content-Type") === "application/json") {
        self.postMessage({ totalTicks: Number(response.headers.get("Content-Length")) });

        let result = "";
        for await (const part of icingaDependencyMapUtils.readStream(response.body)) {
            result += part;
            self.postMessage({ tick: result.length });
        }

        self.postMessage({ networkIdle: true });

        preparer = function* (messages) {
            for (const message of messages) {
                yield message;
            }
        }(JSON.parse(result));
    } else {
        preparer = prepareLayers(...await parseStructure(response.body));
    }

    let next = preparer.next();

    do {
        self.postMessage(next.value);
    } while (! (next = preparer.next()).done);

    if (typeof next.value !== "undefined") {
        const body = JSON.stringify(next.value);

        try {
            const response = await fetch(cacheUrl, {
                method: "POST",
                body,
                headers: {
                    "Content-Type": "application/json",
                    "Content-Length": body.length
                }
            });

            if (response.status !== 200) {
                self.postMessage({
                    success: false,
                    reason: "Failed to cache dependency map structure: " + response.statusText
                });
            } else {
                self.postMessage({ success: true });
            }
        } catch (error) {
            self.postMessage({
                success: false,
                reason: "Failed to cache dependency map structure: " + error.message
            });
        }
    } else {
        self.postMessage({ success: true });
    }
}

self.onmessage = e => {
    const { baseUrl, group, invalidate, csrfToken } = e.data;
    main(csrfToken, { baseUrl, group, invalidate })
        .catch(error => {
            self.postMessage({ error });
        });
};
